from __future__ import annotations

import fnmatch
import importlib
import json
import os
import pathlib
import re
import shutil
import subprocess
import sys
import tempfile

from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING

from bokeh.application.handlers.code_runner import CodeRunner

from ..custom import ReactComponent, ReactiveESM

if TYPE_CHECKING:
    from .custom import ExportSpec

GREEN, RED, RESET = "\033[0;32m", "\033[0;31m", "\033[0m"

# Regex pattern to match import statements with URLs starting with https
_ESM_IMPORT_RE = re.compile(
    r'import\s+.*?\s+from\s+["\']'     # Match 'import ... from' with any content
    r'(https:\/\/[^\/]+\/(?:npm\/)?'   # Match the base URL (e.g., https://cdn.jsdelivr.net/) and ignore /npm if present
    r'((?:@[\w\.\-]+\/)?[\w\.\-]+)'    # Capture the package name, including optional scope
    r'(?:@([\d\.\w-]+))?'              # Optionally capture the version after @
    r'[^"\']*)["\']'                   # Match the rest of the URL up to the quote
)
_ESM_IMPORT_SUFFIX = re.compile(r'\/([^?"&\']*)')

# Regex pattern to extract version specifiers from a URL
_ESM_URL_RE = re.compile(
    r'(https:\/\/[^\/]+\/(?:npm\/)?'
    r'((?:@[\w\.\-]+\/)?[\w\.\-]+)'
    r'(?:@([\d\.\w-]+))?'
    r'[^"\']*)'
)
_ESM_IMPORT_ALIAS_RE = re.compile(r'(import\s+(?:\*\s+as\s+\w+|\{[^}]*\}|[\w*\s,]+)\s+from\s+[\'"])(.*?)([\'"])')
_EXPORT_DEFAULT_RE = re.compile(r'\bexport\s+default\b')


@contextmanager
def setup_build_dir(build_dir: str | os.PathLike | None = None):
    original_directory = os.getcwd()
    if build_dir:
        temp_dir = pathlib.Path(build_dir).absolute()
        temp_dir.mkdir(parents=True, exist_ok=True)
    else:
        temp_dir = tempfile.mkdtemp()
    try:
        os.chdir(temp_dir)
        yield temp_dir
    finally:
        os.chdir(original_directory)
        if not build_dir:
            shutil.rmtree(temp_dir)


def check_cli_tool(tool_name):
    try:
        result = subprocess.run([tool_name, '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.returncode == 0:
            return True
        else:
            return False
    except Exception:
        return False


def find_module_bundles(module_spec: str) -> dict[pathlib.PurePath, list[ReactiveESM]]:
    """
    Takes module specifications and extracts a set of components to bundle.

    Arguments
    ---------
    module_spec: str
         Module specification either as a dotted module or a path to a module.

    Returns
    -------
    Dictionary containing the bundle paths and list of components to bundle.
    """
    # Split module spec, while respecting Windows drive letters
    if ':' in module_spec and (module_spec[1:3] != ':\\' or module_spec.count(':') > 1):
        module, cls = module_spec.rsplit(':', 1)
    else:
        module = module_spec
        cls = ''
    classes = cls.split(',') if cls else None
    if module.endswith('.py'):
        module_name, _ = os.path.splitext(os.path.basename(module))
    else:
        module_name = module
    try:
        components = find_components(module, classes)
    except ValueError:
        cls_error = f' and that class(es) {cls!r} are defined therein' if cls else ''
        raise RuntimeError(  # noqa
            f'Could not find any ESM components to compile, ensure '
            f'you provided the right module{cls_error}.'
        )
    if module in sys.modules:
        module_path = sys.modules[module].__file__
    else:
        module_path = module

    bundles = defaultdict(list)
    module_path = pathlib.Path(module_path).parent
    for component in components:
        if component._bundle:
            bundle_path = component._bundle
            if isinstance(bundle_path, str):
                path = (module_path / bundle_path).absolute()
            else:
                path = bundle_path.absolute()
            bundles[str(path)].append(component)
        elif len(components) > 1 and not classes:
            component_module = module_name or component.__module__
            bundles[module_path / f'{component_module}.bundle.js'].append(component)
        else:
            bundles[component._module_path / f'{component.__name__}.bundle.js'].append(component)

    return bundles


def find_components(module_or_file: str | os.PathLike, classes: list[str] | None = None) -> list[type[ReactiveESM]]:
    """
    Creates a temporary module given a path-like object and finds all
    the ReactiveESM components defined therein.

    Arguments
    ---------
    module_or_file : str | os.PathLike
        The path to the Python module.
    classes: list[str] | None
        Names of classes to return.

    Returns
    -------
    List of ReactiveESM components defined in the module.
    """
    py_file = module_or_file.endswith('.py')
    if py_file:
        path_obj = pathlib.Path(module_or_file)
        source = path_obj.read_text(encoding='utf-8')
        runner = CodeRunner(source, module_or_file, [])
        module = runner.new_module()
        runner.run(module)
        if runner.error:
            raise RuntimeError(
                f'Compilation failed because supplied module errored on import:\n\n{runner.error}'
            )
    else:
        module = importlib.import_module(module_or_file)
    classes = classes or []
    components = []
    for v in module.__dict__.values():
        if (
            isinstance(v, type) and
            issubclass(v, ReactiveESM) and
            not v.abstract and
            (not classes or any(fnmatch.fnmatch(v.__name__, p) for p in classes))
        ):
            if py_file:
                v.__path__ = path_obj.parent.absolute()
            components.append(v)
    not_found = {cls for cls in classes if '*' not in cls} - set(c.__name__ for c in components)
    if classes and not_found:
        clss = ', '.join(map(repr, not_found))
        raise ValueError(f'{clss} class(es) not found in {module_or_file!r}.')
    return components


def packages_from_code(esm_code: str) -> dict[str, str]:
    """
    Extracts package version definitions from ESM code.

    Arguments
    ---------
    esm_code : str
        The ESM code to search for package imports.

    Returns
    -------
    Dictionary of packages and their versions.
    """
    packages = {}
    for match in _ESM_IMPORT_RE.findall(esm_code):
        url, package_name, version = match
        packages[package_name] = f'^{version}'
        after_slash_match = _ESM_IMPORT_SUFFIX.search(url.split('@')[-1])
        import_name = package_name
        if after_slash_match:
            suffix = after_slash_match.group(1)
            if suffix != '+esm' and not suffix.endswith(('.js', '.mjs')):
                # ESM specifier is used by some CDNs to load ESM bundle
                import_name += f'/{suffix}'
        esm_code = esm_code.replace(url, import_name)
    return esm_code, packages


def replace_imports(esm_code: str, replacements: dict[str, str]) -> dict[str, str]:
    """
    Replaces imports in the code which may be aliases with the actual
    package names.

    Arguments
    ---------
    esm_code: str
        The ESM code to replace import names in.
    replacements: dict[str, str]
        Mapping that defines replacements from aliased import names
        to actual package names.

    Returns
    -------
    modified_code: str
        The code where imports have been replaced with package names.
    """

    def replace_match(match):
        import_part = match.group(1)
        module_path = match.group(2)
        quote = match.group(3)
        for old, new in replacements.items():
            if module_path.startswith(old):
                module_path = module_path.replace(old, new, 1)
                break
        return f"{import_part}{module_path}{quote}"

    # Use the sub method to replace the matching parts of the import statements
    modified_code = _ESM_IMPORT_ALIAS_RE.sub(replace_match, esm_code)
    return modified_code


def packages_from_importmap(esm_code: str, imports: dict[str, str]) -> dict[str, str]:
    """
    Extracts package version definitions from an import map.

    Arguments
    ---------
    esm_code: str
        The ESM code to replace import names in.
    imports : dict[str, str]
        A dictionary representing the import map, where keys are package names and values are URLs.

    Returns
    -------
    dict[str, str]
        A dictionary where keys are package names and values are their corresponding versions.
    """
    dependencies, replacements = {}, {}
    for key, url in imports.items():
        match = _ESM_URL_RE.search(url)
        if not match:
            raise RuntimeError(
                f'Could not determine package name from URL: {url!r}.'
            )
        pkg_name = match.group(2)
        version = match.group(3)
        dependencies[pkg_name] = f"^{version}" if version else "latest"
        replacements[key] = pkg_name+'/' if key.endswith('/') else pkg_name
    esm_code = replace_imports(esm_code, replacements)
    return esm_code, dependencies


def extract_dependencies(component: type[ReactiveESM]) -> tuple[str, dict[str, any]]:
    """
    Extracts dependencies from a ReactiveESM component by parsing its
    importmap and the associated code and replaces URL import
    specifiers with package imports.

    Arguments
    ---------
    component: type[ReactiveESM]
        The ReactiveESM component to extract a dependency definition from.

    Returns
    -------
    code: str
        Code where the URL imports have been replaced by package imports.
    dependencies: dict[str, str]
        A dictionary of package dependencies and their versions.
    """
    importmap = component._process_importmap()
    esm = component._render_esm(compiled='compiling')
    esm, dependencies = packages_from_importmap(esm, importmap.get('imports', {}))
    esm, packages = packages_from_code(esm)
    dependencies.update(packages)
    return esm, dependencies


def merge_exports(old: ExportSpec, new: ExportSpec):
    """
    Appends the new exports to set of existing ones.

    Appropriately combines different kinds of exports including
    default, import-all exports and named exports.
    """
    for export, specs in new.items():
        if export not in old:
            old[export] = list(specs)
            continue
        prev = old[export]
        for spec in specs:
            if isinstance(spec, tuple):
                for i, p in enumerate(prev):
                    if isinstance(p, tuple):
                        prev[i] = tuple(dict.fromkeys(p+spec))
                        break
                else:
                    prev.append(spec)
            elif spec not in prev:
                prev.append(spec)


def generate_index(imports: str, exports: list[str], export_spec: ExportSpec):
    index_js = imports
    exports = list(exports)
    for module, specs in export_spec.items():
        for spec in specs:
            # TODO: Handle name clashes in exports
            if isinstance(spec, tuple):
                imports = f"{{{', '.join(spec)}}}"
                exports.extend(spec)
            elif spec.startswith('*'):
                imports = f"* as {spec[1:]}"
                exports.append(spec[1:])
            else:
                imports = spec
                exports.append(spec)
            index_js += f'import {imports} from "{module}"\n'

    export_string = ', '.join(exports)
    index_js += f"export default {{{export_string}}}"
    return index_js


def generate_project(
    components: list[type[ReactiveESM]],
    path: str | os.PathLike,
    project_config: dict[str, any] = None
):
    """
    Converts a set of ESM components into a Javascript project with
    an index.js, package.json and a T|JS(X) per component.
    """
    path = pathlib.Path(path)
    component_names = []
    dependencies, export_spec = {}, {}
    index = ''
    for component in components:
        name = component.__name__
        esm_path = component._esm_path(compiled=False)
        if esm_path:
            ext = esm_path.suffix.lstrip('.')
        else:
            ext = 'jsx' if issubclass(component, ReactComponent) else 'js'
        code, component_deps = extract_dependencies(component)
        # Detect default export in component code and handle import accordingly
        if _EXPORT_DEFAULT_RE.search(code):
            index += f'import {name} from "./{name}"\n'
        else:
            index += f'import * as {name} from "./{name}"\n'

        with open(path / f'{name}.{ext}', 'w') as component_file:
            component_file.write(code)
        # TODO: Improve merging of dependencies
        dependencies.update(component_deps)
        merge_exports(export_spec, component._exports__)
        component_names.append(name)

    # Create package.json and write to temp directory
    package_json = {"dependencies": dependencies}
    if project_config:
        package_json.update(project_config)
    with open(path / 'package.json', 'w') as package_json_file:
        json.dump(package_json, package_json_file, indent=2)

    # Generate index file from component imports, exports and export specs
    index_js = generate_index(index, component_names, export_spec)
    with open(path / 'index.js', 'w') as index_js_file:
        index_js_file.write(index_js)


def compile_components(
    components: list[type[ReactiveESM]],
    build_dir: str | os.PathLike = None,
    outfile: str | os.PathLike = None,
    minify: bool = True,
    verbose: bool = True
) -> str | None:
    """
    Compiles a list of ReactiveESM components into a single JavaScript bundle
    including their Javascript dependencies.

    Arguments
    ---------
    components : list[type[ReactiveESM]]
        A list of `ReactiveESM` component classes to compile.
    build_dir : str | os.PathLike, optional
        The directory where the build output will be saved. If None, a
        temporary directory will be used.
    outfile : str | os.PathLike, optional
        The path to the output file where the compiled bundle will be saved.
        If None the compiled output will be returned.
    minify : bool, optional
        If True, minifies the compiled JavaScript bundle.
    verbose : bool, optional
        If True, prints detailed logs during the compilation process.

    Returns
    -------
    Returns the compiled bundle or None if outfile is provided.
    """
    npm_cmd = 'npm.cmd' if sys.platform == 'win32' else 'npm'
    if not check_cli_tool(npm_cmd):
        raise RuntimeError(
            'Could not find `npm` or it generated an error. Ensure it is '
            'installed and can be run with `npm --version`. You can get it '
            'with conda or you favorite package manager or nodejs manager.'
        )
    if not check_cli_tool('esbuild'):
        raise RuntimeError(
            'Could not find `esbuild` or it generated an error. Ensure it '
            'is installed and can be run with `esbuild --version`. You can '
            'install it with conda or with `npm install -g esbuild`.'
        )

    out = str(pathlib.Path(outfile).absolute()) if outfile else None
    with setup_build_dir(build_dir) as build_dir:
        generate_project(components, build_dir)
        extra_args = []
        if verbose:
            extra_args.append('--log-level=debug')
        install_cmd = [npm_cmd, 'install'] + extra_args
        try:
            if out:
                print(f"Running command: {' '.join(install_cmd)}\n")  # noqa
            result = subprocess.run(install_cmd, check=True, capture_output=True, text=True)
            if result.stdout and out:
                print(f"npm output:\n{GREEN}{result.stdout}{RESET}")  # noqa
            if result.stderr:
                print(f"npm errors:\n{RED}{result.stderr}{RESET}")  # noqa

        except subprocess.CalledProcessError as e:
            print(f"An error occurred while running npm install:\n{RED}{e.stderr}{RESET}")  # noqa
            return None

        if any(issubclass(c, ReactComponent) for c in components):
            extra_args.append('--loader:.js=jsx')
        if minify:
            extra_args.append('--minify')
        if out:
            extra_args.append(f'--outfile={out}')
        build_cmd = ['esbuild', 'index.js', '--bundle', '--format=esm'] + extra_args
        try:
            if verbose:
                print(f"Running command: {' '.join(build_cmd)}\n")  # noqa
            result = subprocess.run(build_cmd+['--color=true'], check=True, capture_output=True, text=True)
            if result.stderr:
                print(f"esbuild output:\n{result.stderr}")  # noqa
                return None
        except subprocess.CalledProcessError as e:
            print(f"An error occurred while running esbuild: {e.stderr}")  # noqa
            return None
        return 0 if outfile else result.stdout



from bokeh.core.properties import Override
from bokeh.models.widgets.sliders import DateSlider


class DatetimeSlider(DateSlider):
    """ Slider-based datetime selection widget. """

    step = Override(default=60)

    format = Override(default="%d %b %Y %H:%M:%S")



import param

import panel as pn

from panel.custom import Child, ReactComponent


class Example(ReactComponent):

    child = Child()

    child2 = Child()

    color = param.Color()

    text = param.String()

    celebrate = param.Boolean()

    _esm = 'react_demo.js'

    _importmap = {
        "imports": {
            "@emotion/cache": "https://esm.sh/@emotion/cache",
            "@emotion/react": "https://esm.sh/@emotion/react@11.11.4",
            "@mui/material/": "https://esm.sh/@mui/material@5.11.10/",
            "canvas-confetti": "https://esm.sh/canvas-confetti@1.6.0",
        },
        "scopes": {
        }
    }

example = Example(text='Hello World!', child='Wow!')

button = pn.widgets.Button(on_click=lambda e: example.param.update(child='Woo!'), name='Update')

pn.Row(
    pn.Param(example.param, parameters=['color', 'text', 'celebrate']),
    example, button
).servable()



import asyncio
import pathlib

import aiohttp

THUMBNAIL_URL = 'https://assets.holoviz.org/panel'
GALLERY_DIR = pathlib.Path(__file__).parent.parent.parent / 'examples' / 'gallery'
THUMBNAIL_DIR = GALLERY_DIR / 'thumbnails'
GALLERY_ITEMS = GALLERY_DIR.glob('*.ipynb')
THUMBNAIL_DIR.mkdir(exist_ok=True)

async def download(path, name, session):
    filename = f'{name[:-6].lower()}.png'
    url = f'{THUMBNAIL_URL}/{path}/{filename}'
    async with session.get(url) as response:
        if response.content_type != 'image/png':
            return
        with open(THUMBNAIL_DIR / filename, "wb") as f:
            f.write(await response.read())

async def download_all():
    async with aiohttp.ClientSession() as session:
        await asyncio.gather(
            *[download(item.parent.name, item.name, session) for item in GALLERY_ITEMS]
        )

asyncio.run(download_all())



from panel.custom import Children, ReactComponent


class Example(ReactComponent):

    children = Children()

    _esm = """
    export function render({ model }) {
      return <div>{model.get_child("children")}</div>
    }"""

Example(children=[
    'A Markdown pane!',
    'Another Markdown pane!'
]).servable()




from unittest.mock import MagicMock, patch

import pytest

pytest.importorskip("langchain", reason="Cannot test PanelCallbackHandler without langchain")

from langchain.agents import AgentType, initialize_agent, load_tools
from langchain.chat_models.openai import ChatOpenAI
from langchain.llms.fake import FakeListLLM

from panel.chat import ChatFeed, ChatInterface
from panel.chat.langchain import PanelCallbackHandler


@pytest.mark.parametrize("streaming", [True, False])
@pytest.mark.parametrize("instance_type", [ChatFeed, ChatInterface])
def test_panel_callback_handler(streaming, instance_type):
    async def callback(contents, user, instance):
        await agent.arun(contents)

    instance = instance_type(callback=callback, callback_user="Langchain")
    callback_handler = PanelCallbackHandler(instance)
    tools = load_tools(["python_repl"])
    responses = ["Action: Python REPL\nAction Input: print(2 + 2)", "Final Answer: 4"]
    llm_kwargs = dict(
        responses=responses, callbacks=[callback_handler], streaming=streaming
    )
    llm = FakeListLLM(**llm_kwargs)
    agent = initialize_agent(
        tools,
        llm,
        agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
        callbacks=[callback_handler],
    )
    instance.send("2 + 2")
    assert len(instance.objects) == 3
    assert (
        instance.objects[1].object == "Action: Python REPL\nAction Input: print(2 + 2)"
    )
    assert instance.objects[2].object == "Final Answer: 4"
    assert not instance.disabled


@pytest.mark.parametrize("instance_type", [ChatFeed, ChatInterface])
def test_chat_model(instance_type, mock_completion):
    async def callback(contents, user, instance):
        with patch.object(
            llm,
            "client",
            mock_client,
        ):
            res = llm.predict(contents)
            assert res == "Bar Baz"

    instance = instance_type(
        callback=callback, callback_user="Langchain", callback_exception="verbose"
    )
    callback_handler = PanelCallbackHandler(instance)
    llm = ChatOpenAI(callbacks=[callback_handler], streaming=False)
    mock_client = MagicMock()
    completed = False

    def mock_create(*args, **kwargs):
        nonlocal completed
        completed = True
        return mock_completion

    mock_client.create = mock_create
    instance.send("bar")
    assert completed
    assert not callback_handler._is_streaming
    callback_message = instance.objects[-1]
    assert callback_message.object == "Bar Baz"
    assert callback_message.user == "LangChain (gpt-3.5-turbo)"
    assert len(instance.objects) == 2


def test_stream():
    instance = ChatFeed(callback_user="Langchain")
    callback_handler = PanelCallbackHandler(instance)
    # send each token / chunk individually
    message = callback_handler._stream("Panel")
    callback_handler._message = message
    callback_handler._stream(" ")  # important it's still sent
    callback_handler._stream("Chat!")
    assert instance.objects[-1].object == "Panel Chat!"



import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.chat import ChatInterface
from panel.tests.util import serve_component

pytestmark = pytest.mark.ui


def test_chat_interface_help(page):
    chat_interface = ChatInterface(
        help_text="This is a test help text"
    )
    serve_component(page, chat_interface)

    expect(page.locator("p")).to_have_text("This is a test help text")


def test_chat_interface_custom_js(page):
    chat_interface = ChatInterface()
    chat_interface.button_properties={
        "help": {
            "icon": "help",
            "js_on_click": {
                "code": "console.log(`Typed: '${chat_input.value}'`)",
                "args": {"chat_input": chat_interface.active_widget},
            },
        },
    }
    serve_component(page, chat_interface)

    chat_input = page.locator(".bk-input")
    chat_input.fill("Hello")

    with page.expect_console_message() as msg_info:
        page.locator("button", has_text="help").click()
        msg = msg_info.value

    assert msg.args[0].json_value() == "Typed: 'Hello'"


def test_chat_interface_custom_js_string(page):
    chat_interface = ChatInterface()
    chat_interface.button_properties={
        "help": {
            "icon": "help",
            "js_on_click": "console.log(`Clicked`)",
        },
    }
    serve_component(page, chat_interface)

    chat_input = page.locator(".bk-input")
    chat_input.fill("Hello")

    with page.expect_console_message() as msg_info:
        page.locator("button", has_text="help").click()
        msg = msg_info.value

    assert msg.args[0].json_value() == "Clicked"



import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.tests.util import serve_component, wait_until
from panel.widgets import Player

pytestmark = pytest.mark.ui


def test_player_faster_click_shows_ms(page):
    player = Player()
    serve_component(page, player)

    faster_element = page.locator(".faster")
    faster_element.click()

    wait_until(lambda: player.interval == 350)
    assert faster_element.inner_text() == "2.9\nfps"

    wait_until(lambda: faster_element.inner_text() == "")


def test_player_slower_click_shows_ms(page):
    player = Player()
    serve_component(page, player)

    slower_element = page.locator(".slower")
    slower_element.click()

    wait_until(lambda: player.interval == 714)
    assert slower_element.inner_text() == "1.4\nfps"

    wait_until(lambda: slower_element.inner_text() == "")


def test_init(page):
    player = Player()
    serve_component(page, player)

    assert not page.is_visible('pn-player-value')
    assert page.query_selector('.pn-player-value') is None


def test_show_value(page):
    player = Player(show_value=True)
    serve_component(page, player)

    wait_until(lambda: page.query_selector('.pn-player-value') is not None)
    assert page.query_selector('.pn-player-value') is not None


def test_name(page):
    player = Player(name='test')
    serve_component(page, player)

    assert page.is_visible('label')
    assert page.query_selector('.pn-player-value') is None

    name = page.locator('.pn-player-title:has-text("test")')
    expect(name).to_have_count(1)


def test_value_align(page):
    player = Player(name='test', value_align='end')
    serve_component(page, player)

    name = page.locator('.pn-player-title:has-text("test")')
    expect(name).to_have_css("text-align", "right")


def test_name_and_show_value(page):
    player = Player(name='test', show_value=True)
    serve_component(page, player)

    assert page.is_visible('label')
    assert page.query_selector('.pn-player-value') is not None

    name = page.locator('.pn-player-title:has-text("test")')
    expect(name).to_have_count(1)
def test_player_visible_buttons(page):
    player = Player(visible_buttons=["play", "pause"])
    serve_component(page, player)

    assert page.is_visible(".play")
    assert page.is_visible(".pause")
    assert not page.is_visible(".reverse")
    assert not page.is_visible(".first")
    assert not page.is_visible(".previous")
    assert not page.is_visible(".next")
    assert not page.is_visible(".last")
    assert not page.is_visible(".slower")
    assert not page.is_visible(".faster")

    player.visible_buttons = ["first"]
    expect(page.locator(".first")).to_be_visible()
    assert not page.is_visible(".play")
    assert not page.is_visible(".pause")


def test_player_visible_loop_options(page):
    player = Player(visible_loop_options=["loop", "once"])
    serve_component(page, player)

    assert page.is_visible(".loop")
    assert page.is_visible(".once")
    assert not page.is_visible(".reflect")

    player.visible_loop_options = ["reflect"]
    expect(page.locator(".reflect")).to_be_visible()
    assert not page.is_visible(".loop")
    assert not page.is_visible(".once")


def test_player_scale_buttons(page):
    player = Player(scale_buttons=2)
    serve_component(page, player)

    expect(page.locator(".play")).to_have_attribute(
        "style",
        "text-align: center; flex-grow: 2; margin: 2px; transform: scale(2); max-width: 50px;",
    )



import asyncio
import time

import pytest

from panel.chat.feed import ChatFeed
from panel.chat.icon import ChatReactionIcons
from panel.chat.message import DEFAULT_AVATARS, ChatMessage
from panel.chat.step import ChatStep
from panel.chat.utils import avatar_lookup
from panel.layout import Column, Row
from panel.pane.image import Image
from panel.pane.markup import HTML
from panel.tests.util import async_wait_until, wait_until
from panel.widgets.indicators import LinearGauge
from panel.widgets.input import TextAreaInput, TextInput

LAYOUT_PARAMETERS = {
    "sizing_mode": "stretch_height",
    "height": 201,
    "max_height": 301,
    "width": 101,
    "max_width": 201,
}

ChatFeed.callback_exception = "raise"


@pytest.fixture
def chat_feed():
    return ChatFeed()


@pytest.mark.xdist_group("chat")
class TestChatFeed:

    def test_init_with_help_text(self):
        chat_feed = ChatFeed(help_text="Instructions")
        message = chat_feed.objects[0]
        assert message.object == "Instructions"
        assert message.user == "Help"

    def test_init_with_loading(self):
        chat_feed = ChatFeed(loading=True)
        assert chat_feed._placeholder in chat_feed._chat_log
        chat_feed.loading = False
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_update_header(self):
        chat_feed = ChatFeed(header="1")
        assert chat_feed._card.header == "1"
        chat_feed.header = "2"
        assert chat_feed._card.header == "2"
        chat_feed.header = HTML("<b>3</b>")
        assert chat_feed._card.header.object == "<b>3</b>"

    def test_hide_header(self, chat_feed):
        assert chat_feed.header is None

        chat_feed.header = "# Header"
        assert not chat_feed._card.hide_header

        chat_feed.header = None
        assert chat_feed._card.hide_header

        chat_feed.header = ""
        assert chat_feed._card.hide_header

    def test_card_params(self, chat_feed):
        chat_feed.card_params = {
            "header_background": "red",
            "header": "Test",
            "hide_header": False
        }
        assert chat_feed._card.header_background == "red"
        assert chat_feed._card.header == "Test"
        assert not chat_feed._card.hide_header

    def test_send(self, chat_feed):
        message = chat_feed.send("Message", footer_objects=[HTML("Footer")])
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Message"
        assert chat_feed.objects[0].footer_objects[0].object == "Footer"

    def test_link_chat_log_objects(self, chat_feed):
        chat_feed.send("Message")
        assert chat_feed._chat_log.objects[0] is chat_feed.objects[0]

    def test_send_with_user_avatar(self, chat_feed):
        user = "Bob"
        avatar = "👨"
        message = chat_feed.send("Message", user=user, avatar=avatar)
        assert message.user == user
        assert message.avatar == avatar

    def test_send_dict(self, chat_feed):
        message = chat_feed.send({"object": "Message", "user": "Bob", "avatar": "👨"})
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Message"
        assert chat_feed.objects[0].user == "Bob"
        assert chat_feed.objects[0].avatar == "👨"

    @pytest.mark.parametrize("key", ["value", "object"])
    def test_send_dict_minimum(self, chat_feed, key):
        message = chat_feed.send({key: "Message"})
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Message"

    def test_send_dict_without_object(self, chat_feed):
        with pytest.raises(ValueError, match="it must contain an 'object' key"):
            chat_feed.send({"user": "Bob", "avatar": "👨"})

    def test_send_dict_with_value_and_object(self, chat_feed):
        with pytest.raises(ValueError, match="both 'value' and 'object'"):
            chat_feed.send({"value": "hey", "object": "hi", "user": "Bob", "avatar": "👨"})

    def test_send_dict_with_user_avatar_override(self, chat_feed):
        user = "August"
        avatar = "👩"
        message = chat_feed.send(
            {"object": "Message", "user": "Bob", "avatar": "👨"},
            user=user,
            avatar=avatar,
        )
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Message"
        assert chat_feed.objects[0].user == user
        assert chat_feed.objects[0].avatar == avatar

    def test_send_entry(self, chat_feed):
        message = ChatMessage("Message", user="Bob", avatar="👨")
        chat_feed.send(message)
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Message"
        assert chat_feed.objects[0].user == "Bob"
        assert chat_feed.objects[0].avatar == "👨"

    def test_send_with_respond(self, chat_feed):
        def callback(contents, user, instance):
            return f"Response to: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Question", respond=True)

        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Response to: Question"

        chat_feed.respond()

        wait_until(lambda: len(chat_feed.objects) == 3)
        assert chat_feed.objects[2].object == "Response to: Response to: Question"

    def test_send_without_respond(self, chat_feed):
        def callback(contents, user, instance):
            return f"Response to: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Question", respond=False)

        wait_until(lambda: len(chat_feed.objects) == 1)

        chat_feed.respond()

        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Response to: Question"

    def test_respond_without_callback(self, chat_feed):
        chat_feed.respond()  # Should not raise any errors

    def test_stream(self, chat_feed):
        message = chat_feed.stream("Streaming message", user="Person", avatar="P", footer_objects=[HTML("Footer")])
        assert len(chat_feed.objects) == 1
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Streaming message"
        assert chat_feed.objects[0].user == "Person"
        assert chat_feed.objects[0].avatar == "P"
        assert chat_feed.objects[0].footer_objects[0].object == "Footer"

        updated_entry = chat_feed.stream(
            " Appended message", user="New Person", message=message, avatar="N", footer_objects=[HTML("New Footer")]
        )
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is updated_entry
        assert chat_feed.objects[0].object == "Streaming message Appended message"
        assert chat_feed.objects[0].user == "New Person"
        assert chat_feed.objects[0].avatar == "N"
        assert chat_feed.objects[0].footer_objects[0].object == "New Footer"

        new_entry = chat_feed.stream("New message")
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1] is new_entry
        assert chat_feed.objects[1].object == "New message"

    def test_add_step(self, chat_feed):
        # new
        with chat_feed.add_step("Object", title="Title") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Title"
            assert step.objects[0].object == "Object"

        assert len(chat_feed) == 1
        message = chat_feed.objects[0]
        assert isinstance(message, ChatMessage)
        assert message.user == "Assistant"

        steps = message.object
        assert isinstance(steps, Column)

        assert len(steps) == 1
        assert isinstance(steps[0], ChatStep)

        # existing
        with chat_feed.add_step("New Object", title="New Title") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "New Title"
            assert step.objects[0].object == "New Object"
        assert len(steps) == 2
        assert isinstance(steps[0], ChatStep)
        assert isinstance(steps[1], ChatStep)

        # actual component
        with chat_feed.add_step("Newest Object", title="Newest Title") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Newest Title"
            assert step.objects[0].object == "Newest Object"
        assert len(steps) == 3
        assert isinstance(steps[0], ChatStep)
        assert isinstance(steps[1], ChatStep)
        assert isinstance(steps[2], ChatStep)

    def test_add_step_new_user(self, chat_feed):
        with chat_feed.add_step("Object", title="Title", user="A") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Title"
            assert step.objects[0].object == "Object"

        with chat_feed.add_step("Object 2", title="Title 2", user="B") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Title 2"
            assert step.objects[0].object == "Object 2"

        assert len(chat_feed) == 2
        message1 = chat_feed.objects[0]
        assert isinstance(message1, ChatMessage)
        assert message1.user == "A"
        steps1 = message1.object
        assert isinstance(steps1, Column)
        assert len(steps1) == 1
        assert isinstance(steps1[0], ChatStep)
        assert len(steps1[0].objects) == 1
        assert steps1[0].objects[0].object == "Object"

        message2 = chat_feed.objects[1]
        assert isinstance(message2, ChatMessage)
        assert message2.user == "B"
        steps2 = message2.object
        assert isinstance(steps2, Column)
        assert len(steps2) == 1
        assert isinstance(steps2[0], ChatStep)
        assert len(steps2[0].objects) == 1
        assert steps2[0].objects[0].object == "Object 2"

    def test_add_step_explict_not_append(self, chat_feed):
        with chat_feed.add_step("Object", title="Title") as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Title"
            assert step.objects[0].object == "Object"

        with chat_feed.add_step("Object 2", title="Title 2", append=False) as step:
            assert isinstance(step, ChatStep)
            assert step.title == "Title 2"
            assert step.objects[0].object == "Object 2"

        assert len(chat_feed) == 2
        message1 = chat_feed.objects[0]
        assert isinstance(message1, ChatMessage)
        assert message1.user == "Assistant"
        steps1 = message1.object
        assert isinstance(steps1, Column)
        assert len(steps1) == 1
        assert isinstance(steps1[0], ChatStep)
        assert len(steps1[0].objects) == 1
        assert steps1[0].objects[0].object == "Object"

        message2 = chat_feed.objects[1]
        assert isinstance(message2, ChatMessage)
        assert message2.user == "Assistant"
        steps2 = message2.object
        assert isinstance(steps2, Column)
        assert len(steps2) == 1
        assert isinstance(steps2[0], ChatStep)
        assert len(steps2[0].objects) == 1
        assert steps2[0].objects[0].object == "Object 2"

    def test_add_step_inherits_callback_exception(self, chat_feed):
        chat_feed.callback_exception = "verbose"
        with chat_feed.add_step("Object", title="Title") as step:
            assert step.callback_exception == "verbose"
            raise ValueError("Testing")
        assert "Traceback" in step.objects[0].object

    def test_stream_with_user_avatar(self, chat_feed):
        user = "Bob"
        avatar = "👨"
        message = chat_feed.stream(
            "Streaming with user and avatar", user=user, avatar=avatar
        )
        assert message.user == user
        assert message.avatar == avatar

    def test_stream_dict(self, chat_feed):
        message = chat_feed.stream(
            {"object": "Streaming message", "user": "Person", "avatar": "P"}
        )
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Streaming message"
        assert chat_feed.objects[0].user == "Person"
        assert chat_feed.objects[0].avatar == "P"

    def test_stream_dict_minimum(self, chat_feed):
        message = chat_feed.stream({"object": "Streaming message"})
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Streaming message"

    def test_stream_dict_without_value(self, chat_feed):
        with pytest.raises(ValueError, match="it must contain an 'object' key"):
            chat_feed.stream({"user": "Person", "avatar": "P"})

    def test_stream_dict_with_user_avatar_override(self, chat_feed):
        user = "Bob"
        avatar = "👨"
        message = chat_feed.stream(
            {"object": "Streaming message", "user": "Person", "avatar": "P"},
            user=user,
            avatar=avatar,
        )
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Streaming message"
        assert chat_feed.objects[0].user == user
        assert chat_feed.objects[0].avatar == avatar

    def test_stream_message(self, chat_feed):
        message = ChatMessage("Streaming message", user="Person", avatar="P")
        chat_feed.stream(message)
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        assert chat_feed.objects[0].object == "Streaming message"
        assert chat_feed.objects[0].user == "Person"
        assert chat_feed.objects[0].avatar == "P"

    def test_stream_message_error_passed_user_avatar(self, chat_feed):
        message = ChatMessage("Streaming message", user="Person", avatar="P")
        with pytest.raises(ValueError, match="Cannot set user or avatar"):
            chat_feed.stream(message, user="Bob", avatar="👨")

    def test_stream_replace(self, chat_feed):
        message = chat_feed.stream("Hello")
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0].object == "Hello"

        message = chat_feed.stream(" World", message=message)
        wait_until(lambda: chat_feed.objects[-1].object == "Hello World")

        chat_feed.stream("Goodbye", message=message, replace=True)
        wait_until(lambda: chat_feed.objects[-1].object == "Goodbye")

    @pytest.mark.parametrize("replace", [True, False])
    def test_stream_originally_none_message(self, chat_feed, replace):
        def callback(contents, user, instance):
            for i in range(3):
                chat_feed.stream(f"{i}.", message=base_message, replace=replace)
        chat_feed.callback = callback
        base_message = ChatMessage()
        chat_feed.send(base_message, respond=True)
        assert chat_feed.objects[0].object == "2." if replace else "0.1.2."

    @pytest.mark.parametrize(
        "obj",
        [
            "Some Text",
            TextInput(value="Some Text"),
            HTML("Some Text"),
            Row(HTML("Some Text")),
        ],
    )
    def test_stream_to_nested_entry(self, chat_feed, obj):
        message = chat_feed.send(
            Row(
                obj,
                Image("https://panel.holoviz.org/_static/logo_horizontal.png"),
            )
        )
        chat_feed.stream(" Added", message=message)
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0] is message
        message_obj = chat_feed.objects[0].object[0]
        if isinstance(message_obj, Row):
            message_obj = message_obj[0]

        if hasattr(message_obj, "object"):
            assert message_obj.object == "Some Text Added"
        elif hasattr(message_obj, "value"):
            assert message_obj.value == "Some Text Added"
        else:
            assert message_obj.objects == "Some Text Added"

    def test_undo(self, chat_feed):
        chat_feed.send("Message 1")
        chat_feed.send("Message 2")
        entry3 = chat_feed.send("Message 3")

        wait_until(lambda: len(chat_feed.objects) == 3)

        undone_entries = chat_feed.undo()
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert undone_entries == [entry3]

        chat_feed.undo(2)
        wait_until(lambda: len(chat_feed.objects) == 0)

    def test_clear(self, chat_feed):
        chat_feed.send("Message 1")
        chat_feed.send("Message 2")

        wait_until(lambda: len(chat_feed.objects) == 2)

        cleared_entries = chat_feed.clear()
        wait_until(lambda: len(chat_feed.objects) == 0)
        assert cleared_entries[0].object == "Message 1"
        assert cleared_entries[1].object == "Message 2"

    def test_set_entries(self, chat_feed):
        chat_feed.send("Message 1")
        chat_feed.send("Message 2")

        wait_until(lambda: len(chat_feed.objects) == 2)

        chat_feed.objects = [ChatMessage("Message 3")]
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0].object == "Message 3"

    @pytest.mark.parametrize(["key", "value"], LAYOUT_PARAMETERS.items())
    def test_layout_parameters_are_propogated_to_card(self, key, value):
        chat_feed = ChatFeed(**{key: value})
        assert getattr(chat_feed, key) == value
        assert getattr(chat_feed._card, key) == value

    def test_width_message_offset_80(self, chat_feed):
        """
        Prevent horizontal scroll bars by subtracting 80px
        which is about the width of the avatar
        and reactions.
        """
        chat_feed.width = 500
        chat_feed.send("Message 1")
        assert chat_feed.objects[0].width == 420

    @pytest.mark.parametrize(
        "user", ["system", "System", " System", " system ", "system-"]
    )
    def test_default_avatars_default(self, chat_feed, user):
        chat_feed.send("Message 1", user=user)

        assert chat_feed.objects[0].user == user
        assert chat_feed.objects[0].avatar == "⚙️"

    def test_default_avatars_superseded_in_dict(self, chat_feed):
        chat_feed.send({"user": "System", "avatar": "👨", "value": "Message 1"})

        assert chat_feed.objects[0].user == "System"
        assert chat_feed.objects[0].avatar == "👨"

    def test_default_avatars_superseded_by_keyword(self, chat_feed):
        chat_feed.send({"user": "System", "value": "Message 1"}, avatar="👨")

        assert chat_feed.objects[0].user == "System"
        assert chat_feed.objects[0].avatar == "👨"

    def test_default_avatars_superseded_in_entry(self, chat_feed):
        chat_feed.send(
            ChatMessage(user="System", avatar="👨", object="Message 1")
        )

        assert chat_feed.objects[0].user == "System"
        assert chat_feed.objects[0].avatar == "👨"

    def test_default_avatars_lookup(self, chat_feed):
        def callback(contents, user, instance):
            yield "Message back"

        chat_feed.callback = callback
        chat_feed.callback_user = "System"
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].user == "System"
        assert chat_feed.objects[1].avatar == avatar_lookup(
            "System",
            None,
            {},
            default_avatars=DEFAULT_AVATARS
        )

    def test_default_avatars_superseded_by_callback_avatar(self, chat_feed):
        def callback(contents, user, instance):
            yield "Message back"

        chat_feed.callback = callback
        chat_feed.callback_user = "System"
        chat_feed.callback_avatar = "S"
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].user == "System"
        assert chat_feed.objects[1].avatar == "S"

    def test_default_avatars_message_params(self, chat_feed):
        chat_feed.message_params["default_avatars"] = {"test1": "1"}
        assert chat_feed.send(value="", user="test1").avatar == "1"

        # has default
        assert chat_feed.send(value="", user="system").avatar == "⚙️"

    def test_no_recursion_error(self, chat_feed):
        chat_feed.send("Some time ago, there was a recursion error like this")

    @pytest.mark.parametrize("callback_avatar", [None, "👨", Image("https://panel.holoviz.org/_static/logo_horizontal.png")])
    def test_callback_avatar(self, chat_feed, callback_avatar):
        def callback(contents, user, instance):
            yield "Message back"

        chat_feed.callback_avatar = callback_avatar
        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].avatar == callback_avatar or "🤖"

    @pytest.mark.asyncio
    async def test_chained_response(self, chat_feed):
        async def callback(contents, user, instance):
            if user == "User":
                yield {
                    "user": "arm",
                    "avatar": "🦾",
                    "value": "Hey, leg! Did you hear the user?",
                }
                instance.respond()
            elif user == "arm":
                for user_entry in instance.objects:
                    if user_entry.user == "User":
                        break
                user_contents = user_entry.object
                yield {
                    "user": "leg",
                    "avatar": "🦿",
                    "value": f'Yeah! They said "{user_contents}".',
                }

        chat_feed.callback = callback
        chat_feed.send("Testing!", user="User")
        await async_wait_until(lambda: len(chat_feed.objects) == 3)
        assert chat_feed.objects[1].user == "arm"
        assert chat_feed.objects[1].avatar == "🦾"
        assert chat_feed.objects[1].object == "Hey, leg! Did you hear the user?"
        assert chat_feed.objects[2].user == "leg"
        assert chat_feed.objects[2].avatar == "🦿"
        assert chat_feed.objects[2].object == 'Yeah! They said "Testing!".'

    def test_respond_callback_returns_none(self, chat_feed):
        def callback(contents, user, instance):
            instance.objects[0].object = "Mutated"

        chat_feed.callback = callback
        chat_feed.send("Testing!", user="User")
        wait_until(lambda: len(chat_feed.objects) == 1)
        assert chat_feed.objects[0].object == "Mutated"

    def test_forward_message_params(self, chat_feed):
        chat_feed = ChatFeed(reaction_icons={"like": "thumb-up"}, reactions=["like"])
        chat_feed.send("Hey!")
        chat_message = chat_feed.objects[0]
        assert chat_feed.message_params == {"reaction_icons": {"like": "thumb-up"}, "reactions": ["like"]}
        assert chat_message.object == "Hey!"
        assert chat_message.reactions == ["like"]
        assert chat_message.reaction_icons.options == {"like": "thumb-up"}

    def test_message_params_no_chat_reaction_icons_instance(self, chat_feed):
        with pytest.raises(ValueError, match="Cannot pass"):
            chat_feed.message_params = {"reaction_icons": ChatReactionIcons(
                options={"like": "thumb-up", "dislike": "thumb-down"})}

    def test_update_chat_log_params(self, chat_feed):
        chat_feed = ChatFeed(load_buffer=5, scroll_button_threshold=5, auto_scroll_limit=5)
        assert chat_feed._chat_log.load_buffer == 5
        assert chat_feed._chat_log.scroll_button_threshold == 5
        assert chat_feed._chat_log.auto_scroll_limit == 5

        chat_feed.load_buffer = 10
        chat_feed.scroll_button_threshold = 10
        chat_feed.auto_scroll_limit = 10
        assert chat_feed._chat_log.load_buffer == 10
        assert chat_feed._chat_log.scroll_button_threshold == 10
        assert chat_feed._chat_log.auto_scroll_limit == 10


@pytest.mark.xdist_group("chat")
class TestChatFeedPromptUser:

    async def test_prompt_user_basic(self, chat_feed):
        text_input = TextInput()

        def callback(component, feed):
            feed.send(component.value)

        async def prompt_and_submit():
            chat_feed.prompt_user(text_input, callback)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)
            text_input.value = "test input"
            submit_button = chat_feed.objects[-1].object[1]
            submit_button.clicks += 1
            await async_wait_until(lambda: len(chat_feed.objects) == 2)

        await asyncio.wait_for(prompt_and_submit(), timeout=5.0)
        assert chat_feed.objects[-1].object == "test input"

    async def test_prompt_user_with_predicate(self, chat_feed):
        text_input = TextInput()

        def predicate(component):
            return len(component.value) > 5

        def callback(component, feed):
            feed.send(component.value)

        async def prompt_and_submit():
            chat_feed.prompt_user(text_input, callback, predicate=predicate)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)

            text_input.value = "short"
            submit_button = chat_feed.objects[-1].object[1]
            assert submit_button.disabled

            text_input.value = "long enough"
            await async_wait_until(lambda: not submit_button.disabled)

            submit_button.clicks += 1
            await async_wait_until(lambda: len(chat_feed.objects) == 2)

        await asyncio.wait_for(prompt_and_submit(), timeout=5.0)
        assert chat_feed.objects[-1].object == "long enough"

    async def test_prompt_user_timeout(self, chat_feed):
        text_input = TextInput()

        def callback(component, feed):
            pytest.fail("Callback should not be called on timeout")

        async def prompt_and_wait():
            chat_feed.prompt_user(text_input, callback, timeout=1)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)
            await async_wait_until(lambda: chat_feed.objects[-1].object[1].disabled)

        await asyncio.wait_for(prompt_and_wait(), timeout=5.0)

        submit_button = chat_feed.objects[-1].object[1]
        assert submit_button.name == "Timed out"
        assert submit_button.button_type == "light"
        assert submit_button.icon == "x"

    async def test_prompt_user_custom_button_params(self, chat_feed):
        text_input = TextInput()

        def callback(component, feed):
            feed.send(component.value)

        custom_button_params = {
            "name": "Custom Submit",
            "button_type": "success",
            "icon": "arrow-right"
        }

        async def prompt_and_check():
            chat_feed.prompt_user(text_input, callback, button_params=custom_button_params)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)

        await asyncio.wait_for(prompt_and_check(), timeout=5.0)

        submit_button = chat_feed.objects[-1].object[1]
        assert submit_button.name == "Custom Submit"
        assert submit_button.button_type == "success"
        assert submit_button.icon == "arrow-right"

    async def test_prompt_user_custom_timeout_button_params(self, chat_feed):
        text_input = TextInput()

        def callback(component, feed):
            pytest.fail("Callback should not be called on timeout")

        custom_timeout_params = {
            "name": "Custom Timeout",
            "button_type": "danger",
            "icon": "alert-triangle"
        }

        async def prompt_and_wait():
            chat_feed.prompt_user(text_input, callback, timeout=1, timeout_button_params=custom_timeout_params)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)
            await async_wait_until(lambda: chat_feed.objects[-1].object[1].disabled)

        await asyncio.wait_for(prompt_and_wait(), timeout=5.0)

        submit_button = chat_feed.objects[-1].object[1]
        assert submit_button.name == "Custom Timeout"
        assert submit_button.button_type == "danger"
        assert submit_button.icon == "alert-triangle"

    async def test_prompt_user_async(self, chat_feed):
        text_input = TextInput()

        async def async_callback(component, feed):
            await asyncio.sleep(0.1)
            feed.send("Callback executed")

        async def prompt_and_submit():
            chat_feed.prompt_user(text_input, async_callback)
            await async_wait_until(lambda: len(chat_feed.objects) == 1)

            submit_button = chat_feed.objects[-1].object[1]
            submit_button.clicks += 1

            await async_wait_until(lambda: len(chat_feed.objects) == 2)

        await asyncio.wait_for(prompt_and_submit(), timeout=5.0)

        assert chat_feed.objects[-1].object == "Callback executed"
        assert chat_feed.objects[-2].object.disabled == True


@pytest.mark.xdist_group("chat")
class TestChatFeedCallback:

    def test_user_avatar(self, chat_feed):
        ChatMessage.default_avatars["bob"] = "👨"

        def echo(contents, user, instance):
            return f"{user}: {contents}"

        chat_feed.callback = echo
        chat_feed.callback_user = "Bob"
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].user == "Bob"
        assert chat_feed.objects[1].avatar == "👨"
        ChatMessage.default_avatars.pop("bob")

    def test_return(self, chat_feed):
        def echo(contents, user, instance):
            return contents

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"

    @pytest.mark.parametrize("callback_user", [None, "Bob"])
    @pytest.mark.parametrize("callback_avatar", [None, "C"])
    def test_return_chat_message(self, chat_feed, callback_user, callback_avatar):
        def echo(contents, user, instance):
            message_kwargs = {}
            if callback_user:
                message_kwargs["user"] = callback_user
            if callback_avatar:
                message_kwargs["avatar"] = callback_avatar
            return ChatMessage(object=contents, **message_kwargs)

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert chat_feed.objects[1].user == callback_avatar or "Assistant"
        assert chat_feed.objects[1].avatar == callback_avatar or "🤖"

    def test_yield(self, chat_feed):
        def echo(contents, user, instance):
            yield contents

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"

    @pytest.mark.asyncio
    async def test_async_return(self, chat_feed):
        async def echo(contents, user, instance):
            return contents

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        await async_wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"

    @pytest.mark.parametrize("callback_user", [None, "Bob"])
    @pytest.mark.parametrize("callback_avatar", [None, "C"])
    def test_yield_chat_message(self, chat_feed, callback_user, callback_avatar):
        def echo(contents, user, instance):
            message_kwargs = {}
            if callback_user:
                message_kwargs["user"] = callback_user
            if callback_avatar:
                message_kwargs["avatar"] = callback_avatar
            yield ChatMessage(object=contents, **message_kwargs)

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert chat_feed.objects[1].user == callback_avatar or "Assistant"
        assert chat_feed.objects[1].avatar == callback_avatar or "🤖"

    @pytest.mark.parametrize("callback_user", [None, "Bob"])
    @pytest.mark.parametrize("callback_avatar", [None, "C"])
    def test_yield_chat_message_stream(self, chat_feed, callback_user, callback_avatar):
        def echo(contents, user, instance):
            message_kwargs = {}
            if callback_user:
                message_kwargs["user"] = callback_user
            if callback_avatar:
                message_kwargs["avatar"] = callback_avatar
            message = ""
            for char in contents:
                message += char
                yield ChatMessage(object=message, **message_kwargs)

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert chat_feed.objects[1].user == callback_avatar or "Assistant"
        assert chat_feed.objects[1].avatar == callback_avatar or "🤖"

    @pytest.mark.asyncio
    async def test_async_yield(self, chat_feed):
        async def echo(contents, user, instance):
            yield contents

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        await async_wait_until(lambda: len(chat_feed.objects) == 2)
        assert len(chat_feed.objects) == 2
        assert chat_feed.objects[1].object == "Message"

    def test_generator(self, chat_feed):
        def echo(contents, user, instance):
            message = ""
            for char in contents:
                message += char
                yield message
                assert instance.objects[-1].show_activity_dot

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert len(chat_feed.objects) == 2
        assert chat_feed.objects[1].object == "Message"
        assert not chat_feed.objects[-1].show_activity_dot

    @pytest.mark.parametrize("key", ["value", "object"])
    def test_generator_dict(self, chat_feed, key):
        def echo(contents, user, instance):
            message = ""
            for char in contents:
                message += char
                yield {key: message}

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert chat_feed.objects[-1].object == "Message"

    @pytest.mark.asyncio
    async def test_async_generator(self, chat_feed):
        async def async_gen(contents):
            for char in contents:
                yield char

        async def echo(contents, user, instance):
            message = ""
            async for char in async_gen(contents):
                message += char
                yield message
                assert instance.objects[-1].show_activity_dot

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        await async_wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert not chat_feed.objects[-1].show_activity_dot

    @pytest.mark.asyncio
    @pytest.mark.parametrize("key", ["value", "object"])
    async def test_async_generator_dict(self, chat_feed, key):
        async def async_gen(contents):
            for char in contents:
                yield char

        async def echo(contents, user, instance):
            message = ""
            async for char in async_gen(contents):
                message += char
                yield {key: message}

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        await async_wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"
        assert chat_feed.objects[-1].object == "Message"

    def test_placeholder_text_params(self, chat_feed):
        def echo(contents, user, instance):
            assert instance._placeholder.user == "Loading..."
            assert instance._placeholder.object == "Thinking..."
            time.sleep(1.25)
            return "hey testing"

        chat_feed.callback = echo
        chat_feed.placeholder_text = "Thinking..."
        chat_feed.placeholder_params = {"user": "Loading..."}
        chat_feed.send("Message", respond=True)

    def test_placeholder_disabled(self, chat_feed):
        def echo(contents, user, instance):
            time.sleep(1.25)
            assert instance._placeholder not in instance._chat_log
            return "hey testing"

        chat_feed.placeholder_threshold = 0
        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_placeholder_enabled(self, chat_feed):
        def echo(contents, user, instance):
            time.sleep(1.25)
            assert instance._placeholder in instance._chat_log
            return chat_feed.stream("hey testing")

        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_placeholder_threshold_under(self, chat_feed):
        async def echo(contents, user, instance):
            await asyncio.sleep(0.25)
            assert instance._placeholder not in instance._chat_log
            return "hey testing"

        chat_feed.placeholder_threshold = 5
        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_placeholder_threshold_under_generator(self, chat_feed):
        async def echo(contents, user, instance):
            assert instance._placeholder not in instance._chat_log
            await asyncio.sleep(0.25)
            assert instance._placeholder not in instance._chat_log
            yield "hey testing"

        chat_feed.placeholder_threshold = 5
        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)

    def test_placeholder_threshold_exceed(self, chat_feed):
        async def echo(contents, user, instance):
            await asyncio.sleep(0.5)
            assert instance._placeholder in instance._chat_log
            return "hello testing"

        chat_feed.placeholder_threshold = 0.1
        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_placeholder_threshold_exceed_generator(self, chat_feed):
        async def echo(contents, user, instance):
            await async_wait_until(lambda: instance._placeholder not in instance._chat_log)
            await asyncio.sleep(0.5)
            await async_wait_until(lambda: instance._placeholder in instance._chat_log)
            yield "hello testing"
            await async_wait_until(lambda: instance._placeholder not in instance._chat_log)

        chat_feed.placeholder_threshold = 1
        chat_feed.callback = echo
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_renderers_pane(self, chat_feed):
        chat_feed.renderers = [HTML]
        chat_feed.send("Hello!")
        html = chat_feed.objects[0]._object_panel
        assert isinstance(html, HTML)
        assert html.object == "Hello!"
        assert html.sizing_mode is None

    def test_renderers_widget(self, chat_feed):
        chat_feed.renderers = [TextAreaInput]
        chat_feed.send("Hello!")
        area_input = chat_feed[0]._update_object_pane()
        area_input = chat_feed[0]._object_panel
        assert isinstance(area_input, TextAreaInput)
        assert area_input.value == "Hello!"
        assert area_input.height == 500
        assert area_input.sizing_mode is None

    def test_renderers_custom_callable(self, chat_feed):
        def renderer(value):
            return Column(value, LinearGauge(value=int(value), width=100))

        chat_feed.renderers = [renderer]
        chat_feed.send(1)
        column = chat_feed.objects[0]._object_panel
        assert isinstance(column, Column)
        number = column[0]
        gauge = column[1]
        assert number.object == 1
        assert number.sizing_mode is None
        assert isinstance(gauge, LinearGauge)
        assert gauge.value == 1
        assert gauge.width == 100
        assert gauge.sizing_mode == "fixed"

    def test_callback_exception(self, chat_feed):
        def callback(msg, user, instance):
            return 1 / 0

        chat_feed.callback = callback
        chat_feed.callback_exception = "summary"
        chat_feed.send("Message", respond=True)
        assert "division by zero" in chat_feed.objects[-1].object
        assert chat_feed.objects[-1].user == "Exception"

    def test_callback_exception_traceback(self, chat_feed):
        def callback(msg, user, instance):
            return 1 / 0

        chat_feed.callback = callback
        chat_feed.callback_exception = "verbose"
        chat_feed.send("Message", respond=True)
        assert chat_feed.objects[-1].object.startswith(
            "```python\nTraceback (most recent call last):"
        )
        assert chat_feed.objects[-1].user == "Exception"

    def test_callback_exception_ignore(self, chat_feed):
        def callback(msg, user, instance):
            return 1 / 0

        chat_feed.callback = callback
        chat_feed.callback_exception = "ignore"
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 1)

    def test_callback_exception_raise(self, chat_feed):
        def callback(msg, user, instance):
            return 1 / 0

        chat_feed.callback = callback
        chat_feed.callback_exception = "raise"
        with pytest.raises(ZeroDivisionError, match="division by zero"):
            chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 1)

    def test_callback_stop_generator(self, chat_feed):
        def callback(msg, user, instance):
            yield "A"
            assert chat_feed.stop()
            time.sleep(0.5)
            yield "B"

        chat_feed.callback = callback
        try:
            chat_feed.send("Message", respond=True)
        except asyncio.CancelledError:  # tests pick up this error
            pass
        # use sleep here instead of wait for because
        # the callback is timed and I want to confirm stop works
        time.sleep(1)
        assert chat_feed.objects[-1].object == "A"

    def test_callback_stop_async_generator(self, chat_feed):
        async def callback(msg, user, instance):
            yield "A"
            assert chat_feed.stop()
            await asyncio.sleep(0.5)
            yield "B"

        chat_feed.callback = callback
        try:
            chat_feed.send("Message", respond=True)
        except asyncio.CancelledError:  # tests pick up this error
            pass
        # use sleep here instead of wait for because
        # the callback is timed and I want to confirm stop works
        time.sleep(1)
        assert chat_feed.objects[-1].object == "A"

    def test_callback_stop_function(self, chat_feed):
        def callback(msg, user, instance):
            assert chat_feed.stop()
            return "B"

        chat_feed.callback = callback
        try:
            chat_feed.send("Message", respond=True)
        except asyncio.CancelledError:  # tests pick up this error
            pass
        assert chat_feed.objects[-1].object == "Message"

    def test_callback_stop_async_function(self, chat_feed):
        async def callback(msg, user, instance):
            message = instance.stream("A")
            assert chat_feed.stop()
            await asyncio.sleep(0.5)
            instance.stream("B", message=message)

        chat_feed.callback = callback
        try:
            chat_feed.send("Message", respond=True)
        except asyncio.CancelledError:
            pass
        # use sleep here instead of wait for because
        # the callback is timed and I want to confirm stop works
        time.sleep(1)
        assert chat_feed.objects[-1].object == "A"

    def test_callback_short_time(self, chat_feed):
        def callback(contents, user, instance):
            time.sleep(1)
            message = None
            string = ""
            for c in "helloooo":
                string += c
                time.sleep(0.001)
                message = instance.stream(string, message=message, replace=True)

        feed = ChatFeed(callback=callback)
        feed.send("Message", respond=True)
        assert feed.objects[-1].object == "helloooo"
        assert chat_feed._placeholder not in chat_feed._chat_log

    def test_callback_one_argument(self, chat_feed):
        def callback(contents):
            return contents

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "Message"

    def test_callback_positional_argument(self, chat_feed):
        def callback(*args):
            return f"{args[1]}: {args[0]}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_mix_positional_argument(self, chat_feed):
        def callback(contents, *args):
            return f"{args[0]}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_keyword_argument(self, chat_feed):
        def callback(**kwargs):
            assert "instance" in kwargs
            return f"{kwargs['user']}: {kwargs['contents']}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_mix_keyword_argument(self, chat_feed):
        def callback(contents, **kwargs):
            return f"{kwargs['user']}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_mix_positional_keyword_argument(self, chat_feed):
        def callback(*args, **kwargs):
            assert not kwargs
            return f"{args[1]}: {args[0]}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_two_arguments(self, chat_feed):
        def callback(contents, user):
            return f"{user}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_two_arguments_with_keyword(self, chat_feed):
        def callback(contents, user=None):
            return f"{user}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_three_arguments_with_keyword(self, chat_feed):
        def callback(contents, user=None, instance=None):
            return f"{user}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_two_arguments_yield(self, chat_feed):
        def callback(contents, user):
            yield f"{user}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_two_arguments_async_yield(self, chat_feed):
        async def callback(contents, user):
            yield f"{user}: {contents}"

        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_as_method(self, chat_feed):
        class Test:
            def callback(self, contents, user):
                return f"{user}: {contents}"

        chat_feed.callback = Test().callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_callback_as_class_method(self, chat_feed):
        class Test:
            @classmethod
            def callback(cls, contents, user):
                return f"{user}: {contents}"

        chat_feed.callback = Test.callback
        chat_feed.send("Message", respond=True)
        wait_until(lambda: len(chat_feed.objects) == 2)
        assert chat_feed.objects[1].object == "User: Message"

    def test_persist_placeholder_while_loading(self, chat_feed):
        def callback(contents):
            assert chat_feed._placeholder in chat_feed._chat_log
            return "hey testing"

        chat_feed.loading = True
        chat_feed.callback = callback
        chat_feed.send("Message", respond=True)
        assert chat_feed._placeholder in chat_feed._chat_log


@pytest.mark.xdist_group("chat")
class TestChatFeedSerializeForTransformers:

    def test_defaults(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="user")
        chat_feed.send("I'm the assistant", user="assistant")
        chat_feed.send("I'm a bot", user="bot")
        assert chat_feed.serialize() == [
            {"role": "user", "content": "I'm a user"},
            {"role": "assistant", "content": "I'm the assistant"},
            {"role": "assistant", "content": "I'm a bot"},
        ]

    def test_empty(self):
        chat_feed = ChatFeed()
        assert chat_feed.serialize() == []

    def test_case_insensitivity(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="USER")
        chat_feed.send("I'm the assistant", user="ASSISTant")
        chat_feed.send("I'm a bot", user="boT")
        assert chat_feed.serialize() == [
            {"role": "user", "content": "I'm a user"},
            {"role": "assistant", "content": "I'm the assistant"},
            {"role": "assistant", "content": "I'm a bot"},
        ]

    def test_default_role(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="user")
        chat_feed.send("I'm the assistant", user="assistant")
        chat_feed.send("I'm a bot", user="bot")
        assert chat_feed.serialize(default_role="system") == [
            {"role": "user", "content": "I'm a user"},
            {"role": "assistant", "content": "I'm the assistant"},
            {"role": "system", "content": "I'm a bot"},
        ]

    def test_empty_default_role(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="user")
        chat_feed.send("I'm the assistant", user="assistant")
        chat_feed.send("I'm a bot", user="bot")
        with pytest.raises(ValueError, match="not found in role_names"):
            chat_feed.serialize(default_role="")

    def test_role_names(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm the user", user="Andrew")
        chat_feed.send("I'm another user", user="August")
        chat_feed.send("I'm the assistant", user="Bot")
        role_names = {"user": ["Andrew", "August"], "assistant": "Bot"}
        assert chat_feed.serialize(role_names=role_names) == [
            {"role": "user", "content": "I'm the user"},
            {"role": "user", "content": "I'm another user"},
            {"role": "assistant", "content": "I'm the assistant"},
        ]

    def test_custom_serializer(self):
        def custom_serializer(obj):
            if isinstance(obj, str):
                return "new string"
            else:
                return "0"

        chat_feed = ChatFeed()
        chat_feed.send("I'm the user", user="user")
        chat_feed.send(3, user="assistant")
        assert chat_feed.serialize(custom_serializer=custom_serializer) == [
            {"role": "user", "content": "new string"},
            {"role": "assistant", "content": "0"},
        ]

    def test_custom_serializer_invalid_output(self):
        def custom_serializer(obj):
            if isinstance(obj, str):
                return "new string"
            else:
                return 0

        chat_feed = ChatFeed()
        chat_feed.send("I'm the user", user="user")
        chat_feed.send(3, user="assistant")
        with pytest.raises(ValueError, match="must return a string"):
            chat_feed.serialize(custom_serializer=custom_serializer)

    def test_serialize_filter_by(self, chat_feed):
        def filter_by_reactions(messages):
            return [obj for obj in messages if "favorite" in obj.reactions]

        chat_feed.send(ChatMessage("no"))
        chat_feed.send(ChatMessage("yes", reactions=["favorite"]))
        filtered = chat_feed.serialize(filter_by=filter_by_reactions)
        assert len(filtered) == 1
        assert filtered[0]["content"] == "yes"

    def test_serialize_exclude_users_default(self):
        def say_hi(contents, user, instance):
            return f"Hi {user}!"

        chat_feed = ChatFeed(
            help_text="This chat feed will respond by saying hi!",
            callback=say_hi
        )
        chat_feed.send("Hello there!")
        assert chat_feed.serialize() == [
            {"role": "user", "content": "Hello there!"},
            {"role": "assistant", "content": "Hi User!"}
        ]

    def test_serialize_exclude_users_custom(self):
        def say_hi(contents, user, instance):
            return f"Hi {user}!"

        chat_feed = ChatFeed(
            help_text="This chat feed will respond by saying hi!",
            callback=say_hi
        )
        chat_feed.send("Hello there!")
        assert chat_feed.serialize(exclude_users=["assistant"]) == [
            {"role": "assistant", "content": "This chat feed will respond by saying hi!"},
            {"role": "user", "content": "Hello there!"},
        ]

    def test_serialize_exclude_placeholder(self):
        def say_hi(contents, user, instance):
            assert len(instance.serialize()) == 1
            return f"Hi {user}!"

        chat_feed = ChatFeed(
            help_text="This chat feed will respond by saying hi!",
            callback=say_hi
        )

        chat_feed.send("Hello there!")
        assert chat_feed.serialize() == [
            {"role": "user", "content": "Hello there!"},
            {"role": "assistant", "content": "Hi User!"}
        ]

    def test_serialize_limit(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="user")
        chat_feed.send("I'm the assistant", user="assistant")
        chat_feed.send("I'm a bot", user="bot")
        assert chat_feed.serialize(limit=1) == [
            {"role": "assistant", "content": "I'm a bot"},
        ]

    def test_serialize_class(self, chat_feed):
        class Test():

            def __repr__(self):
                return "Test()"

        chat_feed.send(Test())
        assert chat_feed.serialize() == [{"role": "user", "content": "Test()"}]

    def test_serialize_kwargs(self, chat_feed):
        chat_feed.send("Hello")
        chat_feed.add_step("Hello", "World")
        assert chat_feed.serialize(
            prefix_with_container_label=False,
            prefix_with_viewable_label=False
        ) == [
            {'role': 'user', 'content': 'Hello'},
            {'role': 'assistant', 'content': '((Hello))'}
        ]


@pytest.mark.xdist_group("chat")
class TestChatFeedSerializeBase:

    def test_transformers_format(self):
        chat_feed = ChatFeed()
        chat_feed.send("I'm a user", user="user")
        chat_feed.send("I'm the assistant", user="assistant")
        chat_feed.send("I'm a bot", user="bot")
        assert chat_feed.serialize(format="transformers") == [
            {"role": "user", "content": "I'm a user"},
            {"role": "assistant", "content": "I'm the assistant"},
            {"role": "assistant", "content": "I'm a bot"},
        ]

    def test_invalid(self):
        with pytest.raises(NotImplementedError, match="is not supported"):
            chat_feed = ChatFeed()
            chat_feed.send("I'm a user", user="user")
            chat_feed.serialize(format="atransform")


@pytest.mark.xdist_group("chat")
class TestChatFeedPostHook:

    def test_return_string(self, chat_feed):
        def callback(contents, user, instance):
            yield f"Echo: {contents}"

        def append_callback(message, instance):
            logs.append(message.object)

        logs = []
        chat_feed.callback = callback
        chat_feed.post_hook = append_callback
        chat_feed.send("Hello World!")
        wait_until(lambda: chat_feed.objects[-1].object == "Echo: Hello World!")
        assert logs == ["Hello World!", "Echo: Hello World!"]

    def test_yield_string(self, chat_feed):
        def callback(contents, user, instance):
            yield f"Echo: {contents}"

        def append_callback(message, instance):
            logs.append(message.object)

        logs = []
        chat_feed.callback = callback
        chat_feed.post_hook = append_callback
        chat_feed.send("Hello World!")
        wait_until(lambda: chat_feed.objects[-1].object == "Echo: Hello World!")
        assert logs == ["Hello World!", "Echo: Hello World!"]

    def test_generator(self, chat_feed):
        def callback(contents, user, instance):
            message = "Echo: "
            for char in contents:
                message += char
                yield message

        def append_callback(message, instance):
            logs.append(message.object)

        logs = []
        chat_feed.callback = callback
        chat_feed.post_hook = append_callback
        chat_feed.send("Hello World!")
        wait_until(lambda: chat_feed.objects[-1].object == "Echo: Hello World!")
        assert logs == ["Hello World!", "Echo: Hello World!"]

    def test_async_generator(self, chat_feed):
        async def callback(contents, user, instance):
            message = "Echo: "
            for char in contents:
                message += char
                yield message

        async def append_callback(message, instance):
            logs.append(message.object)

        logs = []
        chat_feed.callback = callback
        chat_feed.post_hook = append_callback
        chat_feed.send("Hello World!")
        wait_until(lambda: chat_feed.objects[-1].object == "Echo: Hello World!")
        assert logs == ["Hello World!", "Echo: Hello World!"]

    def test_stream(self, chat_feed):
        def callback(contents, user, instance):
            message = instance.stream("Echo: ")
            for char in contents:
                message = instance.stream(char, message=message)

        def append_callback(message, instance):
            logs.append(message.object)

        logs = []
        chat_feed.callback = callback
        chat_feed.post_hook = append_callback
        chat_feed.send("AB")
        wait_until(lambda: chat_feed.objects[-1].object == "Echo: AB")
        assert logs == ["AB", "Echo: ", "Echo: AB"]



"""
Extensions for Bokeh application handling.
"""
from __future__ import annotations

import json
import logging
import os

from functools import partial
from types import FunctionType, MethodType
from typing import (
    TYPE_CHECKING, Any, Callable, Mapping,
)
from urllib.parse import urljoin

import bokeh.command.util

from bokeh.application import Application as BkApplication
from bokeh.application.handlers.directory import DirectoryHandler
from bokeh.application.handlers.document_lifecycle import (
    DocumentLifecycleHandler,
)
from bokeh.application.handlers.function import FunctionHandler
from bokeh.models import CustomJS

from ..config import config
from .document import _destroy_document
from .handlers import MarkdownHandler, NotebookHandler, ScriptHandler
from .loading import LOADING_INDICATOR_CSS_CLASS
from .logging import LOG_SESSION_DESTROYED, LOG_SESSION_LAUNCHING
from .state import set_curdoc, state

if TYPE_CHECKING:
    from bokeh.application.application import SessionContext
    from bokeh.application.handlers.handler import Handler
    from bokeh.document.document import Document

    from ..template.base import BaseTemplate
    from ..viewable import Viewable, Viewer
    from .location import Location

    TViewable = Viewable | Viewer | BaseTemplate
    TViewableFuncOrPath = TViewable | Callable[[], TViewable] | os.PathLike | str


logger = logging.getLogger('panel.io.application')


def _eval_panel(
    panel: TViewableFuncOrPath, server_id: str, title: str,
    location: bool | Location, admin: bool, doc: Document
):
    from ..pane import panel as as_panel
    from ..template import BaseTemplate

    if config.global_loading_spinner:
        doc.js_on_event(
            'document_ready', CustomJS(code=f"""
            const body = document.getElementsByTagName('body')[0]
            body.classList.remove({LOADING_INDICATOR_CSS_CLASS!r}, {config.loading_spinner!r})
            """)
        )

    doc.on_event('document_ready', partial(state._schedule_on_load, doc))

    # Set up instrumentation for logging sessions
    logger.info(LOG_SESSION_LAUNCHING, id(doc))
    def _log_session_destroyed(session_context):
        logger.info(LOG_SESSION_DESTROYED, id(doc))
    doc.on_session_destroyed(_log_session_destroyed)

    with set_curdoc(doc):
        if isinstance(panel, (FunctionType, MethodType)):
            panel = panel()
        if isinstance(panel, BaseTemplate):
            doc = panel._modify_doc(server_id, title, doc, location)
        else:
            doc = as_panel(panel)._modify_doc(server_id, title, doc, location)
        return doc

def _on_session_destroyed(session_context: SessionContext) -> None:
    """
    Calls any on_session_destroyed callbacks defined on the Document
    """
    callbacks = session_context._document.session_destroyed_callbacks
    session_context._document.session_destroyed_callbacks = set()
    for callback in callbacks:
        try:
            callback(session_context)
        except Exception as e:
            logger.warning(
                "DocumentLifecycleHandler on_session_destroyed "
                f"callback {callback} failed with following error: {e}"
            )


class Application(BkApplication):
    """
    Extends Bokeh Application with ability to add global session
    creation callbacks, support for the admin dashboard and the
    ability to globally define a template.
    """

    def __init__(self, *args, **kwargs):
        self._admin = kwargs.pop('admin', None)
        super().__init__(*args, **kwargs)

    async def on_session_created(self, session_context):
        with set_curdoc(session_context._document):
            if self._admin is not None:
                config._admin = self._admin
            for cb in state._on_session_created_internal+state._on_session_created:
                cb(session_context)
        await super().on_session_created(session_context)

    def add(self, handler: Handler) -> None:
        """
        Override default DocumentLifeCycleHandler
        """
        if type(handler) is DocumentLifecycleHandler:
            handler._on_session_destroyed = _on_session_destroyed
        super().add(handler)

    def _set_session_prefix(self, doc):
        session_context = doc.session_context
        if not (session_context and session_context.server_context):
            return
        request = session_context.request
        app_context = session_context.server_context.application_context
        prefix = request.uri.replace(app_context._url, '')
        if not prefix.endswith('/'):
            prefix += '/'
        base_url = urljoin('/', prefix)
        rel_path = '/'.join(['..'] * app_context._url.strip('/').count('/'))

        # Handle autoload.js absolute paths
        abs_url = request.arguments.get('bokeh-absolute-url')
        if abs_url:
            rel_path = abs_url[0].decode('utf-8').replace(app_context._url, '')

        with set_curdoc(doc):
            state.base_url = base_url
            state.rel_path = rel_path

    def initialize_document(self, doc):
        logger.info(LOG_SESSION_LAUNCHING, id(doc))
        self._set_session_prefix(doc)
        super().initialize_document(doc)
        if doc in state._templates and doc not in state._templates[doc]._documents:
            template = state._templates[doc]
            with set_curdoc(doc):
                template.server_doc(title=template.title, location=True, doc=doc)
        def _log_session_destroyed(session_context):
            logger.info(LOG_SESSION_DESTROYED, id(doc))
        doc.destroy = partial(_destroy_document, doc) # type: ignore
        doc.on_event('document_ready', partial(state._schedule_on_load, doc))
        doc.on_session_destroyed(_log_session_destroyed)

    def process_request(self, request) -> dict[str, Any]:
        ''' Processes incoming HTTP request returning a dictionary of
        additional data to add to the session_context.

        Args:
            request: HTTP request

        Returns:
            A dictionary of JSON serializable data to be included on
            the session context.
        '''
        request_data = super().process_request(request)
        user = request.cookies.get('user')
        if user and config.cookie_secret:
            from tornado.web import decode_signed_value
            try:
                user = decode_signed_value(config.cookie_secret, 'user', user.value).decode('utf-8')
            except Exception:
                user = user.value
            if user in state._oauth_user_overrides:
                user_data = json.dumps(state._oauth_user_overrides[user])
                if state.encryption:
                    user_data = state.encryption.encrypt(user_data.encode('utf-8'))
                request_data['user_data'] = user_data
        return request_data

bokeh.command.util.Application = Application # type: ignore


def build_single_handler_application(path, argv=None):
    argv = argv or []
    path = os.path.abspath(os.path.expanduser(path))
    handler: Handler

    # There are certainly race conditions here if the file/directory is deleted
    # in between the isdir/isfile tests and subsequent code. But it would be a
    # failure if they were not there to begin with, too (just a different error)
    if os.path.isdir(path):
        handler = DirectoryHandler(filename=path, argv=argv)
    elif os.path.isfile(path):
        if path.endswith(".ipynb"):
            handler = NotebookHandler(filename=path, argv=argv)
        elif path.endswith(".md"):
            handler = MarkdownHandler(filename=path, argv=argv)
        elif path.endswith(".py"):
            handler = ScriptHandler(filename=path, argv=argv)
        else:
            raise ValueError(f"Expected a '.py' script or '.ipynb' notebook, got: {path!r}" % path)
    else:
        raise ValueError(f"Path for Bokeh server application does not exist: {path}")

    if handler.failed and not config.autoreload:
        raise RuntimeError(f"Error loading {path}:\n\n{handler.error}\n{handler.error_detail} ")

    application = Application(handler)

    return application

bokeh.command.util.build_single_handler_application = build_single_handler_application


def build_applications(
    panel: TViewableFuncOrPath | Mapping[str, TViewableFuncOrPath],
    title: str | dict[str, str] | None = None,
    location: bool | Location = True,
    admin: bool = False,
    server_id: str | None = None,
    custom_handlers: list | None = None
) -> dict[str, Application]:
    """
    Converts a variety of objects into a dictionary of Applications.

    Arguments
    ---------
    panel: Viewable, function or {str: Viewable}
        A Panel object, a function returning a Panel object or a
        dictionary mapping from the URL slug to either.
    title : str or {str: str} (optional, default=None)
        An HTML title for the application or a dictionary mapping
        from the URL slug to a customized title.
    location : boolean or panel.io.location.Location
        Whether to create a Location component to observe and
        set the URL location.
    admin: boolean (default=False)
        Whether to enable the admin panel
    server_id: str
        ID of the server running the application(s)
    """
    if not isinstance(panel, dict):
        panel = {'/': panel}

    apps = {}
    for slug, app in panel.items():
        if slug.endswith('/') and slug != '/':
            raise ValueError(f"Invalid URL: trailing slash '/' used for {slug!r} not supported.")
        if isinstance(title, dict):
            try:
                title_ = title[slug]
            except KeyError:
                raise KeyError(
                    "Keys of the title dictionary and of the apps "
                    f"dictionary must match. No {slug} key found in the "
                    "title dictionary.") from None
        else:
            title_ = title
        slug = slug if slug.startswith('/') else '/'+slug

        # Handle other types of apps using a custom handler
        for handler in (custom_handlers or ()):
            new_app = handler(slug, app)
            if app is not None:
                break
        else:
            new_app = app
        if new_app is not None:
            app = new_app
            if app is True:
                continue

        if isinstance(app, os.PathLike):
            app = str(app) # enables serving apps from Paths
        if (isinstance(app, str) and app.endswith(('.py', '.ipynb', '.md'))
            and os.path.isfile(app)):
            apps[slug] = app = build_single_handler_application(app)
            app._admin = admin
        elif isinstance(app, BkApplication):
            apps[slug] = app
        else:
            handler = FunctionHandler(partial(_eval_panel, app, server_id, title_, location, admin))
            apps[slug] = Application(handler, admin=admin)

    if admin:
        if '/admin' in apps:
            raise ValueError(
                'Cannot enable admin panel because another app is being served '
                'on the /admin endpoint'
            )
        from .admin import admin_panel
        admin_handler = FunctionHandler(admin_panel)
        apps['/admin'] = Application(admin_handler)

    return apps



import datetime

import pytest

from panel.tests.util import serve_component, wait_until
from panel.widgets import TimePicker

pytestmark = pytest.mark.ui


def test_time_picker(page):

    time_picker = TimePicker(value="18:08", format="H:i")

    serve_component(page, time_picker)

    # test init corrected timezone
    locator = page.locator("#input")
    assert locator.get_attribute("value") == "18:08:00"

    # test UI change
    locator = page.locator("input.bk-input.form-control.input")
    locator.click()
    wait_until(lambda: page.locator("input.numInput.flatpickr-hour").is_visible())
    locator = page.locator("input.numInput.flatpickr-hour")
    locator.press("ArrowDown")
    locator.press("Enter")
    wait_until(lambda: time_picker.value == datetime.time(17, 8))

    # test str value change
    time_picker.value = "04:08"
    wait_until(lambda: time_picker.value == "04:08")
    locator = page.locator("#input")
    assert locator.get_attribute("value") == "04:08:00"

    # test datetime.time value change
    time_picker.value = datetime.time(18, 8)
    wait_until(lambda: time_picker.value == datetime.time(18, 8))
    locator = page.locator("#input")
    assert locator.get_attribute("value") == "18:08:00"


@pytest.mark.parametrize("timezone_id", [
    "America/New_York",
    "Europe/Berlin",
    "UTC",
])
def test_time_picker_timezone_different(page, timezone_id):
    context = page.context.browser.new_context(
        timezone_id=timezone_id,
    )
    page = context.new_page()

    time_picker = TimePicker(value="18:08", format="H:i")
    serve_component(page, time_picker)

    locator = page.locator("#input")
    assert locator.get_attribute("value") == "18:08:00"



"""
Custom bokeh Widget models.
"""
from bokeh.core.enums import ButtonType
from bokeh.core.properties import (
    Any, Bool, Either, Enum, Float, Instance, Int, List, Nullable, Override,
    String, Tuple,
)
from bokeh.events import ModelEvent
from bokeh.models.ui import Tooltip
from bokeh.models.ui.icons import Icon
from bokeh.models.widgets import (
    Button as bkButton, CheckboxButtonGroup as bkCheckboxButtonGroup,
    InputWidget, MultiSelect, RadioButtonGroup as bkRadioButtonGroup, Select,
    TextAreaInput as bkTextAreaInput, TextInput as bkTextInput, Widget,
)

from .layout import HTMLBox


class DoubleClickEvent(ModelEvent):

    event_name = 'dblclick_event'

    def __init__(self, model, option=None):
        self.option = option
        super().__init__(model=model)


class Player(Widget):
    """
    The Player widget provides controls to play through a number of frames.
    """
    title = Nullable(String, default="", help="""
    The slider's label (supports :ref:`math text <ug_styling_mathtext>`).
    """)

    start = Int(0, help="Lower bound of the Player slider")

    end = Int(10, help="Upper bound of the Player slider")

    value = Int(0, help="Current value of the player app")

    value_throttled = Int(0, help="Current throttled value of the player app")

    value_align = String("start", help="""Location to display
        the value of the slider ("start" "center", "end")""")

    step = Int(1, help="Number of steps to advance the player by.")

    interval = Int(500, help="Interval between updates")

    direction = Int(0, help="""
        Current play direction of the Player (-1: playing in reverse,
        0: paused, 1: playing)""")

    loop_policy = Enum('once', 'reflect', 'loop', default='once')

    show_loop_controls = Bool(True, help="""Whether the loop controls
        radio buttons are shown""")

    preview_duration = Int(1500, help="""
        Duration (in milliseconds) for showing the current FPS when clicking
        the slower/faster buttons, before reverting to the icon.""")

    show_value = Bool(True, help="""
        Whether to show the widget value""")

    width = Override(default=400)

    height = Override(default=250)

    scale_buttons = Float(1, help="Percentage to scale the size of the buttons by")

    visible_buttons = List(String, default=[
        'slower', 'first', 'previous', 'reverse', 'pause', 'play', 'next', 'last', 'faster'
    ], help="The buttons to display on the player.")

    visible_loop_options = List(String, default=[
        'once', 'loop', 'reflect'
    ], help="The loop options to display on the player.")

class DiscretePlayer(Player):

    options = List(Any, help="""
        List of discrete options.""")


class SingleSelect(InputWidget):
    ''' Single-select widget.

    '''

    disabled_options = List(Any, default=[], help="""
    List of options to disable.
    """)

    options = List(Either(String, Tuple(String, String)), help="""
    Available selection options. Options may be provided either as a list of
    possible string values, or as a list of tuples, each of the form
    ``(value, label)``. In the latter case, the visible widget text for each
    value will be corresponding given label.
    """)

    size = Int(default=4, help="""
    The number of visible options in the dropdown list. (This uses the
    ``select`` HTML element's ``size`` attribute. Some browsers might not
    show less than 3 options.)
    """)

    value = Nullable(String, help="Initial or selected value.")


class Audio(HTMLBox):

    loop = Bool(False, help="""Whether the audio should loop""")

    paused = Bool(False, help="""Whether the audio is paused""")

    muted = Bool(False, help="""Whether the audio is muted""")

    autoplay = Bool(False, help="""Whether the audio is playing automatically""")

    time = Float(0, help="""
        The current time stamp of the audio playback""")

    throttle = Int(250, help="""
        The frequency at which the time value is updated in milliseconds.""")

    value = Any(help="Encoded file data")

    volume = Nullable(Float, help="""The volume of the audio player.""")


class Video(HTMLBox):

    loop = Bool(False, help="""Whether the video should loop""")

    paused = Bool(False, help="""Whether the video is paused""")

    muted = Bool(False, help="""Whether the video is muted""")

    autoplay = Bool(False, help="""Whether the video is playing automatically""")

    time = Float(0, help="""
        The current time stamp of the video playback""")

    throttle = Int(250, help="""
        The frequency at which the time value is updated in milliseconds.""")

    value = String(help="Encoded file data")

    volume = Int(help="""The volume of the video player.""")


class VideoStream(HTMLBox):

    format = Enum('png', 'jpeg', default='png')

    paused = Bool(False, help="""Whether the video is paused""")

    snapshot = Bool(False, help="""On change generate a snapshot of the current video frame""")

    timeout = Nullable(Int, help="""
        The timeout between snapshots (if None snapshot only generated
        when snapshot property is changed""")

    value = Any(help="""Snapshot Data""")

    height = Override(default=240)

    width = Override(default=320)


class Progress(HTMLBox):

    active = Bool(True, help="""Whether to animate the bar""")

    bar_color = Enum('primary', 'secondary', 'success', 'info',
                     'danger', 'warning', 'light', 'dark', default='primary')

    max = Int(100, help="""Maximum value""")

    value = Nullable(Int, help="""Current value""")

    css = List(String)


class FileDownload(InputWidget):

    auto = Bool(False, help="""Whether to download on click""")

    button_type = Enum(ButtonType, help="""
    A style for the button, signifying it's role.
    """)

    clicks = Int(0, help="""
    A private property that used to trigger ``on_click`` event handler.
    """)

    data = String(help="""Encoded URI data.""")

    embed = Bool(False, help="""Whether the data is pre-embedded.""")

    icon = Nullable(Instance(Icon), help="""
    An optional image appearing to the left of button's text. An instance of
    :class:`~bokeh.models.Icon` (such as :class:`~bokeh.models.BuiltinIcon`,
    :class:`~bokeh.models.SVGIcon`, or :class:`~bokeh.models.TablerIcon`).`
    """)

    label = String("", help="""The text label for the button to display.""")

    filename = String(help="""Filename to use on download""")

    _transfers = Int(0, help="""
    A private property to create and click the link.
    """)

    title = Override(default='')


class CustomSelect(Select):
    ''' Custom widget that extends the base Bokeh Select
    by adding a parameter to disable one or more options.

    '''
    disabled_options = List(Any, default=[], help="""
    List of options to disable.
    """)

    size = Int(default=1)


class CustomMultiSelect(MultiSelect):
    """
    MultiSelect widget which allows capturing double tap events.
    """

class TooltipIcon(Widget):
    description = Instance(
        Tooltip,
        default=Tooltip(content="Help text", position="right"),
        help="""The tooltip held by the icon"""
    )


class TextAreaInput(bkTextAreaInput):

    auto_grow = Bool(
        default=False,
        help="""
        Whether the text area should automatically grow vertically to
        accommodate the current text."""
    )

    max_rows = Nullable(Int(), help="""
        Maximum number of rows the input area can grow to if auto_grow
        is enabled."""
    )


class Button(bkButton):

    tooltip = Nullable(Instance(Tooltip), help="""
    A tooltip with plain text or rich HTML contents, providing general help or
    description of a widget's or component's function.
    """)

    tooltip_delay = Int(500, help="""
    Delay (in milliseconds) to display the tooltip after the cursor has
    hovered over the Button, default is 500ms.
    """)


class CheckboxButtonGroup(bkCheckboxButtonGroup):

    tooltip = Nullable(Instance(Tooltip), help="""
    A tooltip with plain text or rich HTML contents, providing general help or
    description of a widget's or component's function.
    """)

    tooltip_delay = Int(500, help="""
    Delay (in milliseconds) to display the tooltip after the cursor has
    hovered over the Button, default is 500ms.
    """)


class RadioButtonGroup(bkRadioButtonGroup):

    tooltip = Nullable(Instance(Tooltip), help="""
    A tooltip with plain text or rich HTML contents, providing general help or
    description of a widget's or component's function.
    """)

    tooltip_delay = Int(500, help="""
    Delay (in milliseconds) to display the tooltip after the cursor has
    hovered over the Button, default is 500ms.
    """)


class EnterEvent(ModelEvent):

    event_name = 'enter-pressed'

    def __init__(self, model, value_input):
        self.value_input = value_input
        super().__init__(model=model)

    def __repr__(self):
        return (
            f'{type(self).__name__}(value_input={self.value_input})'
        )

class TextInput(bkTextInput): ...



"""
Defines the Placeholder pane which serves as a placeholder for other Panel components.
"""

from __future__ import annotations

import param

from ..pane.base import ReplacementPane


class Placeholder(ReplacementPane):
    """
    The `Placeholder` pane serves as a placeholder for other Panel components.
    It can be used to display a message while a computation is running, for
    example.

    Reference: https://panel.holoviz.org/reference/panes/Placeholder.html

    :Example:

    >>> with Placeholder("⏳ Idle"):
    ...     placeholder.object = "🏃 Running..."
    """

    def __init__(self, object=None, **params):
        super().__init__(object=object, **params)
        self._past_object = object  # used to restore object when Placeholder is exited
        self._temporary = False
        if object is not None:
            self._replace_panel()

    @param.depends("object", watch=True)
    def _replace_panel(self):
        if not self._temporary:
            self._past_object = self.object
        self._update_inner(self.object)

    def __enter__(self):
        self._temporary = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.object = self._past_object
        finally:
            self._temporary = False
        return False

    def update(self, object):
        """
        Updates the object on the Placeholder.

        Arguments
        ---------
        object: The object to update the Placeholder with.
        """
        self.object = object



from bokeh.core.properties import (
    Bool, Dict, Enum, Int, List, Nullable, String,
)
from bokeh.events import ModelEvent
from bokeh.models.widgets import InputWidget

from ..config import config
from ..io.resources import bundled_files
from ..util import classproperty


class UploadEvent(ModelEvent):

    event_name = 'upload_event'

    def __init__(self, model, data=None):
        self.data = data
        super().__init__(model=model)

class DeleteEvent(ModelEvent):

    event_name = 'delete_event'

    def __init__(self, model, data=None):
        self.data = data
        super().__init__(model=model)


class FileDropper(InputWidget):

    accepted_filetypes = List(String)

    chunk_size = Int(10_000_000)

    max_files = Nullable(Int)

    max_file_size = Nullable(String)

    max_total_file_size = Nullable(String)

    mime_type = Dict(String, String)

    multiple = Bool(True)

    layout = Nullable(Enum("integrated", "compact", "circle", default="compact"))

    __javascript_raw__ = [
        f"{config.npm_cdn}/filepond-plugin-image-preview/dist/filepond-plugin-image-preview.js",
        f"{config.npm_cdn}/filepond-plugin-file-validate-size/dist/filepond-plugin-file-validate-size.js",
        f"{config.npm_cdn}/filepond-plugin-file-validate-type/dist/filepond-plugin-file-validate-type.js",
        f"{config.npm_cdn}/filepond-plugin-pdf-preview/dist/filepond-plugin-pdf-preview.min.js",
        f"{config.npm_cdn}/filepond@^4/dist/filepond.min.js"
    ]

    @classproperty
    def __javascript__(cls):
        return bundled_files(cls)

    __js_require__ = {
        'paths': {
            "filepond":  f"{config.npm_cdn}/filepond@^4/dist/filepond",
            "filepond-preview-image": f"{config.npm_cdn}/filepond-plugin-image-preview/dist/filepond-plugin-image-preview",
            "filepond-validate-size": f"{config.npm_cdn}/filepond-plugin-file-validate-size/dist/filepond-plugin-file-validate-size",
            "filepond-validate-type": f"{config.npm_cdn}/filepond-plugin-file-validate-type/dist/filepond-plugin-file-validate-type",
            "filepond-preview-pdf": f"{config.npm_cdn}/filepond-plugin-pdf-preview/dist/filepond-plugin-pdf-preview.min",
        },
        'exports': {
            'filepond': 'FilePond',
            'filepond-preview-image': 'FilePondPluginImagePreview',
            'filepond-preview-pdf': 'FilePondPluginPdfPreview',
            'filepond-validate-size': 'FilePondPluginFileValidateSize',
            'filepond-validate-type': 'FilePondPluginFileValidateType'
        }
    }

    @classproperty
    def __js_skip__(cls):
        return {
            'FilePond': cls.__javascript__[:]
        }

    __css_raw__ = [
        f"{config.npm_cdn}/filepond@^4/dist/filepond.css",
        f"{config.npm_cdn}/filepond-plugin-image-preview/dist/filepond-plugin-image-preview.css",
        f"{config.npm_cdn}/filepond-plugin-pdf-preview/dist/filepond-plugin-pdf-preview.css"
    ]

    @classproperty
    def __css__(cls):
        return bundled_files(cls, 'css')



collect_ignore_glob = [
    "apps/",
    "developer_guide/",
    "homepage.ipynb",
    "*VTK*.ipynb",
    "*Vega.ipynb",
    "*DeckGL*.ipynb",
    "*Terminal.ipynb",
]


def pytest_runtest_makereport(item, call):
    """
    Skip tests that fail because "the kernel died before replying to kernel_info"
    this is a common error when running the example tests in CI.

    Inspired from: https://stackoverflow.com/questions/32451811

    """
    from _pytest.runner import pytest_runtest_makereport

    tr = pytest_runtest_makereport(item, call)

    if call.excinfo is not None:
        msgs = [
            "Kernel died before replying to kernel_info",
            "Kernel didn't respond in 60 seconds",
        ]
        for msg in msgs:
            if call.excinfo.type == RuntimeError and call.excinfo.value.args[0] in msg:
                tr.outcome = "skipped"
                tr.wasxfail = f"reason: {msg}"

    return tr



import pathlib

import param
import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.config import config
from panel.custom import (
    AnyWidgetComponent, Child, Children, JSComponent, ReactComponent,
)
from panel.io.compile import compile_components
from panel.layout import Row
from panel.layout.base import ListLike
from panel.pane import Markdown
from panel.tests.util import serve_component, wait_until

pytestmark = pytest.mark.ui


class JSUpdate(JSComponent):

    text = param.String()

    _esm = """
    export function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = model.text
      model.on('text', () => {
        h1.textContent = model.text;
      })
      return h1
    }
    """

class ReactUpdate(ReactComponent):

    text = param.String()

    _esm = """
    export function render({ model }) {
      const [text, setText ] = model.useState("text")
      return <h1>{text}</h1>
    }
    """

class AnyWidgetUpdate(AnyWidgetComponent):

    text = param.String()

    _esm = """
    export function render({ model, el }) {
      const h1 = document.createElement('h1')
      h1.textContent = model.get("text")
      model.on("change:text", () => {
        h1.textContent = model.get("text");
      })
      el.append(h1)
    }
    """

@pytest.mark.parametrize('component', [JSUpdate, ReactUpdate, AnyWidgetUpdate])
def test_update(page, component):
    example = component(text='Hello World!')

    serve_component(page, example)

    expect(page.locator('h1')).to_have_text('Hello World!')

    example.text = "Foo!"

    expect(page.locator('h1')).to_have_text('Foo!')


class AnyWidgetInitialize(AnyWidgetComponent):

    count = param.Integer(default=0)

    _esm = """
    export function initialize({ model }) {
      model.set('count', 1)
      model.save_changes()
    }

    export function render({ model, el }) {
      const h1 = document.createElement('h1')
      h1.textContent = `${model.get('count')}`
      el.append(h1)
    }
    """

class JSInitialize(JSComponent):

    count = param.Integer(default=0)

    _esm = """
    export function initialize({ model }) {
      model.count = 1
    }

    export function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = `${model.count}`
      return h1
    }
    """

class ReactInitialize(ReactComponent):

    count = param.Integer(default=0)

    _esm = """
    export function initialize({ model }) {
      model.count = 1
    }

    export function render({ model }) {
      const [count] = model.useState('count')
      return <h1>{count}</h1>
    }
    """

@pytest.mark.parametrize('component', [AnyWidgetInitialize, JSInitialize, ReactInitialize])
def test_initialize(page, component):
    example = Row(component())

    serve_component(page, example)

    expect(page.locator('h1')).to_have_text('1')

    example[0] = component()

    expect(page.locator('h1')).to_have_text('1')


class AnyWidgetModuleCached(AnyWidgetComponent):

    count = param.Integer(default=0)

    _esm = """
    let count = 0

    export function initialize({ model }) {
      count += 1
      model.set('count', count)
      model.save_changes()
    }

    export function render({ model, el }) {
      const h1 = document.createElement('h1')
      h1.textContent = `${model.get('count')}`
      el.append(h1)
    }
    """

class JSModuleCached(JSComponent):

    count = param.Integer(default=0)

    _esm = """
    let count = 0

    export function initialize({ model }) {
      count += 1
      model.count = count
    }

    export function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = `${model.count}`
      return h1
    }
    """

class ReactModuleCached(ReactComponent):

    count = param.Integer(default=0)

    _esm = """
    let count = 0

    export function initialize({ model }) {
      count += 1
      model.count = count
    }

    export function render({ model }) {
      const [count] = model.useState('count')
      return <h1>{count}</h1>
    }
    """

@pytest.mark.parametrize('component', [AnyWidgetModuleCached, JSModuleCached, ReactModuleCached])
def test_module_cached(page, component):
    example = Row(component())

    serve_component(page, example)

    expect(page.locator('h1')).to_have_text('1')

    example[0] = component()

    expect(page.locator('h1')).to_have_text('2')


class JSUnwatch(JSComponent):

    text = param.String()

    _esm = """
    export function render({ model, el }) {
      const h1 = document.createElement('h1')
      const h2 = document.createElement('h2')
      h1.textContent = model.text
      h2.textContent = model.text
      const cb = () => {
        h1.textContent = model.text;
      }
      const cb2 = () => {
        h2.textContent = model.text;
        model.off('text', cb2)
      }
      model.on('text', cb)
      model.on('text', cb2)
      el.append(h1, h2)
    }
    """

class AnyWidgetUnwatch(AnyWidgetComponent):

    text = param.String()

    _esm = """
    export function render({ model, el }) {
      const h1 = document.createElement('h1')
      const h2 = document.createElement('h2')
      h1.textContent = model.get("text")
      h2.textContent = model.get("text")
      const cb = () => {
        h1.textContent = model.get("text");
      }
      const cb2 = () => {
        h2.textContent = model.get("text");
        model.off("change:text", cb2)
      }
      model.on("change:text", cb)
      model.on("change:text", cb2)
      el.append(h1, h2)
    }
    """

@pytest.mark.parametrize('component', [JSUnwatch, AnyWidgetUnwatch])
def test_unwatch(page, component):
    example = component(text='Hello World!')

    serve_component(page, example)

    expect(page.locator('h1')).to_have_text('Hello World!')
    expect(page.locator('h1')).to_have_text('Hello World!')

    example.text = "Foo!"

    expect(page.locator('h1')).to_have_text('Foo!')
    expect(page.locator('h2')).to_have_text('Foo!')

    example.text = "Baz!"

    expect(page.locator('h1')).to_have_text('Baz!')
    expect(page.locator('h2')).to_have_text('Foo!')


class JSInput(JSComponent):

    text = param.String()

    _esm = """
    export function render({ model }) {
      const inp = document.createElement('input')
      inp.id = 'input'
      inp.value = model.text
      inp.addEventListener('change', (event) => {
        model.text = event.target.value;
      })
      return inp
    }
    """

class ReactInput(ReactComponent):

    text = param.String()

    _esm = """
    export function render({ model }) {
      const [text, setText ] = model.useState("text")
      return (
        <input
          id="input"
          value={text}
          onChange={e => setText(e.target.value)}
        />
      )
    }
    """

@pytest.mark.parametrize('component', [JSInput, ReactInput])
def test_gather_input(page, component):
    example = component(text='Hello World!')

    serve_component(page, example)

    inp = page.locator('#input')

    inp.click()
    for _ in example.text:
        inp.press('Backspace')
    inp.press_sequentially('Foo!')
    inp.press('Enter')

    wait_until(lambda: example.text == 'Foo!', page)


class JSSendEvent(JSComponent):

    clicks = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      const button = document.createElement('button')
      button.id = 'button'
      button.onclick = (event) => model.send_event('click', event)
      return button
    }
    """

    def _handle_click(self, event):
        self.clicks += 1

class ReactSendEvent(ReactComponent):

    clicks = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      return <button id="button" onClick={(event) => model.send_event('click', event)}/>
    }
    """

    def _handle_click(self, event):
        self.clicks += 1

@pytest.mark.parametrize('component', [JSSendEvent, ReactSendEvent])
def test_send_event(page, component):
    button = component()

    serve_component(page, button)

    page.locator('#button').click()

    wait_until(lambda: button.clicks, page)


class JSSendMsg(JSComponent):

    clicks = param.Integer(default=0)

    msg = param.String()

    _esm = """
    export function render({ model }) {
      const button = document.createElement('button')
      button.id = 'button'
      button.onclick = (event) => model.send_msg('click')
      return button
    }
    """

    def _handle_msg(self, msg):
        self.msg = msg
        self.clicks += 1


class ReactSendMsg(ReactComponent):

    clicks = param.Integer(default=0)

    msg = param.String()

    _esm = """
    export function render({ model }) {
      return <button id="button" onClick={(event) => model.send_msg('click')}/>
    }
    """

    def _handle_msg(self, msg):
        self.msg = msg
        self.clicks += 1


@pytest.mark.parametrize('component', [JSSendMsg, ReactSendMsg])
def test_send_msg(page, component):
    button = component()

    serve_component(page, button)

    page.locator('#button').click()

    wait_until(lambda: button.clicks, page)
    assert button.msg == 'click'


class JSChild(JSComponent):

    child = Child()

    render_count = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      const button = document.createElement('button')
      button.appendChild(model.get_child('child'))
      model.render_count += 1
      return button
    }"""

class ReactChild(ReactComponent):

    child = Child()

    render_count = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      model.render_count += 1
      return <button>{model.get_child('child')}</button>
    }"""

@pytest.mark.parametrize('component', [JSChild, ReactChild])
def test_child(page, component):
    example = component(child='A Markdown pane!')

    serve_component(page, example)

    expect(page.locator('button')).to_have_text('A Markdown pane!')

    example.child = 'A different Markdown pane!'

    expect(page.locator('button')).to_have_text('A different Markdown pane!')

    wait_until(lambda: example.render_count == (2 if component is JSChild else 1), page)


class JSChildren(ListLike, JSComponent):

    objects = Children()

    render_count = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      const div = document.createElement('div')
      div.id = "container"
      div.append(...model.get_child('objects'))
      model.render_count += 1
      return div
    }"""

class JSChildrenNoReturn(JSChildren):

    _esm = """
    export function render({ model, view }) {
      const div = document.createElement('div')
      div.id = "container"
      div.append(...model.get_child('objects'))
      view.container.replaceChildren(div)
      model.render_count += 1
    }"""

class ReactChildren(ListLike, ReactComponent):

    objects = Children()

    render_count = param.Integer(default=0)

    _esm = """
    export function render({ model }) {
      model.render_count += 1
      return <div id="container">{model.get_child("objects")}</div>
    }"""

@pytest.mark.parametrize('component', [JSChildren, JSChildrenNoReturn, ReactChildren])
def test_children(page, component):
    example = component(objects=['A Markdown pane!'])

    serve_component(page, example)

    expect(page.locator('#container')).to_have_text('A Markdown pane!')

    example.objects = ['A different Markdown pane!']

    expect(page.locator('#container')).to_have_text('A different Markdown pane!')

    example.objects = ['<div class="foo">1</div>', '<div class="foo">2</div>']

    expect(page.locator('.foo').nth(0)).to_have_text('1')
    expect(page.locator('.foo').nth(1)).to_have_text('2')

    page.wait_for_timeout(400)

    assert example.render_count == (3 if issubclass(component, JSChildren) else 2)

@pytest.mark.parametrize('component', [JSChildren, JSChildrenNoReturn, ReactChildren])
def test_children_add_and_remove_without_error(page, component):
    example = component(objects=['A Markdown pane!'])

    msgs, _ = serve_component(page, example)

    expect(page.locator('#container')).to_have_text('A Markdown pane!')

    example.append('A different Markdown pane!')
    example.pop(-1)

    expect(page.locator('#container')).to_have_text('A Markdown pane!')

    expect(page.locator('.markdown')).to_have_count(1)

    page.wait_for_timeout(500)

    assert [msg for msg in msgs if msg.type == 'error' and 'favicon' not in msg.location['url']] == []

@pytest.mark.parametrize('component', [JSChildren, JSChildrenNoReturn, ReactChildren])
def test_children_append_without_rerender(page, component):
    child = JSChild(child=Markdown(
        'A Markdown pane!', css_classes=['first']
    ))
    example = component(objects=[child])

    serve_component(page, example)

    expect(page.locator('.first')).to_have_text('A Markdown pane!')

    wait_until(lambda: child.render_count == 1, page)

    example.objects = example.objects+[Markdown(
        'A different Markdown pane!', css_classes=['second']
    )]

    expect(page.locator('.second')).to_have_text('A different Markdown pane!')

    page.wait_for_timeout(400)

    assert child.render_count == 1
    assert example.render_count == 2


JS_CODE_BEFORE = """
export function render() {
  const h1 = document.createElement('h1')
  h1.innerText = "foo"
  return h1
}"""

JS_CODE_AFTER = """
export function render() {
  const h1 = document.createElement('h1')
  h1.innerText = "bar"
  return h1
}"""

REACT_CODE_BEFORE = """
export function render() {
  return <h1>foo</h1>
}"""

REACT_CODE_AFTER = """
export function render() {
  return <h1>bar</h1>
}"""

@pytest.mark.parametrize('component,before,after', [
    (JSComponent, JS_CODE_BEFORE, JS_CODE_AFTER),
    (ReactChildren, REACT_CODE_BEFORE, REACT_CODE_AFTER),
])
def test_reload(page, js_file, component, before, after):
    js_file.file.write(before)
    js_file.file.flush()
    js_file.file.seek(0)

    class CustomReload(component):
        _esm = pathlib.Path(js_file.name)

    example = CustomReload()

    with config.set(autoreload=True):
        serve_component(page, example)

        expect(page.locator('h1')).to_have_text('foo')

        js_file.file.write(after)
        js_file.file.flush()
        js_file.file.seek(0)
        example._update_esm()

        expect(page.locator('h1')).to_have_text('bar')


def test_anywidget_custom_event(page):

    class SendEvent(AnyWidgetComponent):

        _esm = """
        export function render({model, el}) {
          const h1 = document.createElement('h1')
          model.on("msg:custom", (msg) => { h1.innerText = msg.text })
          el.append(h1)
        }
        """

    example = SendEvent()
    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    example.send({"type": "foo", "text": "bar"})

    expect(page.locator('h1')).to_have_text("bar")


class JSLifecycleAfterRender(JSComponent):

    _esm = """
    export function render({ model }) {
      const h1 = document.createElement('h1')
      model.on('after_render', () => { h1.textContent = 'rendered' })
      return h1
    }"""

class ReactLifecycleAfterRender(ReactComponent):

    _esm = """
    import {useState} from "react"

    export function render({ model }) {
      const [text, setText] = useState("")
      model.on('after_render', () => { setText('rendered') })
      return <h1>{text}</h1>
    }"""

@pytest.mark.parametrize('component', [JSLifecycleAfterRender, ReactLifecycleAfterRender])
def test_after_render_lifecycle_hooks(page, component):
    example = component()

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("rendered")


class JSLifecycleAfterLayout(JSComponent):

    _esm = """
    export function render({ model }) {
      const h1 = document.createElement('h1')
      model.on('after_layout', () => { h1.textContent = 'layouted' })
      return h1
    }"""

class ReactLifecycleAfterLayout(ReactComponent):

    _esm = """
    import {useState} from "react"

    export function render({ model }) {
      const [text, setText] = useState("")
      model.on('after_layout', () => { setText('layouted') })
      return <h1>{text}</h1>
    }"""

@pytest.mark.parametrize('component', [JSLifecycleAfterLayout, ReactLifecycleAfterLayout])
def test_after_layout_lifecycle_hooks(page, component):
    example = component()

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("layouted")


class JSLifecycleAfterResize(JSComponent):

    _esm = """
    export function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = "0"
      let count = 0
      model.on('resize', () => { count += 1; h1.textContent = `${count}`; })
      return h1
    }"""

class ReactLifecycleAfterResize(ReactComponent):

    _esm = """
    import {useState} from "react"

    export function render({ model }) {
      const [count, setCount] = useState(0)
      model.on('resize', () => { setCount(count+1); })
      return <h1>{count}</h1>
    }"""

@pytest.mark.parametrize('component', [JSLifecycleAfterResize, ReactLifecycleAfterResize])
def test_after_resize_lifecycle_hooks(page, component):
    example = component(sizing_mode='stretch_width')

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("1")

    page.set_viewport_size({ "width": 50, "height": 300})

    expect(page.locator('h1')).to_have_text("2")


class JSLifecycleRemove(JSComponent):

    _esm = """
    export function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = 'Hello'
      model.on('remove', () => { console.warn('Removed') })
      return h1
    }"""

class ReactLifecycleRemove(ReactComponent):

    _esm = """
    import {useState} from "react"

    export function render({ model }) {
      const [count, setCount] = useState(0)
      model.on('remove', () => { console.warn('Removed') })
      return <h1>Hello</h1>
    }"""


@pytest.mark.parametrize('component', [JSLifecycleRemove, ReactLifecycleRemove])
def test_remove_lifecycle_hooks(page, component):
    example = Row(component(sizing_mode='stretch_width'))

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("Hello")

    with page.expect_console_message() as msg_info:
        example.clear()

    wait_until(lambda: msg_info.value.args[0].json_value() == "Removed", page)


class JSDefaultExport(JSComponent):

    _esm = """
    function render({ model }) {
      const h1 = document.createElement('h1')
      h1.textContent = 'Hello'
      return h1
    }

    export default {render}
    """

class AnyWidgetDefaultExport(AnyWidgetComponent):

    _esm = """
    function render({ model, el }) {
      const h1 = document.createElement('h1')
      h1.textContent = 'Hello'
      el.append(h1)
    }

    export default {render}
    """

class ReactDefaultExport(ReactComponent):

    _esm = """
    function render({ model }) {
      return <h1>Hello</h1>
    }

    export default { render }
    """

@pytest.mark.parametrize('component', [AnyWidgetDefaultExport, JSDefaultExport, ReactDefaultExport])
def test_esm_component_default_export(page, component):
    example = Row(component(sizing_mode='stretch_width'))

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("Hello")


class JSDefaultFunctionExport(JSComponent):

    _esm = """
    export default () => {
      function render({ model }) {
        const h1 = document.createElement('h1')
        h1.textContent = 'Hello'
        return h1
      }

      return {render}
    }
    """

class AnyWidgetDefaultFunctionExport(AnyWidgetComponent):

    _esm = """
    export default () => {
      function render({ model, el }) {
        const h1 = document.createElement('h1')
        h1.textContent = 'Hello'
        el.append(h1)
      }

      return {render}
    }
    """

class ReactDefaultFunctionExport(ReactComponent):

    _esm = """
    export default () => {
      function render({ model }) {
        return <h1>Hello</h1>
      }
      return {render}
    }
    """

@pytest.mark.parametrize('component', [AnyWidgetDefaultFunctionExport, JSDefaultExport, ReactDefaultExport])
def test_esm_component_default_function_export(page, component):
    example = Row(component(sizing_mode='stretch_width'))

    serve_component(page, example)

    expect(page.locator('h1')).to_have_count(1)

    expect(page.locator('h1')).to_have_text("Hello")


@pytest.mark.parametrize('component', [AnyWidgetInitialize, JSInitialize, ReactInitialize])
def test_esm_compile_simple(page, component):
    outfile = pathlib.Path(__file__).parent / f'{component.__name__}.bundle.js'
    ret = compile_components([component], outfile=outfile)
    if ret or not outfile.is_file():
        raise RuntimeError('Could not compile ESM component')

    assert component._bundle_path == outfile

    example = Row(component())

    serve_component(page, example)

    expect(page.locator('h1')).to_have_text('1')

    example[0] = component()

    expect(page.locator('h1')).to_have_text('1')


class JSBase(JSComponent):

    _bundle = 'js.bundle.js'

    _esm = """
    export function render({model}) {
      const h1 = document.createElement('h1')
      h1.id = model.name
      h1.textContent = "Rendered"
      return h1
    }
    """

class JS1(JSBase):
    pass

class JS2(JSBase):
    pass


class AnyWidgetBase(AnyWidgetComponent):

    _bundle = 'anywidget.bundle.js'

    _esm = """
    export function render({model, el}) {
      const h1 = document.createElement('h1')
      h1.id = model.get("name")
      h1.textContent = "Rendered"
      el.append(h1)
    }
    """

class AnyWidget1(AnyWidgetBase):
    pass

class AnyWidget2(AnyWidgetBase):
    pass


class ReactBase(ReactComponent):

    _bundle = 'react.bundle.js'

    _esm = """
    export function render({model, el}) {
      return <h1 id={model.name}>Rendered</h1>
    }
    """

class React1(ReactBase):
    pass

class React2(ReactBase):
    pass

@pytest.mark.parametrize('components', [[JS1, JS2], [AnyWidget1, AnyWidget2], [React1, React2]])
def test_esm_compile_shared(page, components):
    component1, component2 = components
    outfile = pathlib.Path(__file__).parent / component1._bundle
    ret = compile_components([component1, component2], outfile=outfile)
    if ret or not outfile.is_file():
        raise RuntimeError('Could not compile ESM component')

    assert component1._bundle_path == outfile
    assert component2._bundle_path == outfile

    example = Row(component1(), component2())

    serve_component(page, example)

    expect(page.locator(f'#{example[0].name}')).to_have_text('Rendered')
    expect(page.locator(f'#{example[1].name}')).to_have_text('Rendered')



import pytest

from scripts.test_panelite.utils import MaxRetriesExceeded, Retrier


def test_retrier_no_exception():
    retrier = Retrier(retries=3)
    while not retrier.accomplished:
        with retrier:
            pass

    assert retrier.retries==0

def test_retrier_exception():
    retrier = Retrier(retries=2, delay=0)
    with pytest.raises(MaxRetriesExceeded):
        while not retrier.accomplished:
            with retrier:
                raise ValueError

    assert retrier.retries==2



import os
import time

import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.tests.util import (
    linux_only, run_panel_serve, wait_for_port, write_file,
)

pytestmark = pytest.mark.ui


@linux_only
def test_autoreload_app(py_file, port, page):
    app = "import panel as pn; pn.Row('Example 1').servable()"
    app2 = "import panel as pn; pn.Row('Example 2').servable()"
    write_file(app, py_file.file)

    app_name = os.path.basename(py_file.name)[:-3]

    with run_panel_serve(["--port", str(port), '--autoreload', py_file.name]) as p:
        port = wait_for_port(p.stdout)
        time.sleep(0.2)

        page.goto(f"http://localhost:{port}/{app_name}")
        expect(page.locator(".markdown")).to_have_text("Example 1")

        # Timeout to ensure websocket is initialized
        time.sleep(1.0)

        write_file(app2, py_file.file)
        expect(page.locator(".markdown")).to_have_text('Example 2')



from __future__ import annotations

import re

from collections.abc import Iterable
from io import BytesIO
from textwrap import indent
from typing import Any, Union

import param

from panel.pane import HTML, Image

from ..io.resources import CDN_DIST
from ..pane.image import ImageBase
from ..viewable import Viewable

Avatar = Union[str, BytesIO, bytes, ImageBase]
AvatarDict = dict[str, Avatar]


def to_alpha_numeric(user: str) -> str:
    """
    Convert the user name to an alpha numeric string,
    removing all non-alphanumeric characters.
    """
    return re.sub(r"\W+", "", user).lower()


def avatar_lookup(
    user: str,
    avatar: Any,
    avatars: dict[str, Any],
    default_avatars: dict[str, Any],
) -> Avatar:
    """
    Lookup the avatar for the user.
    """
    alpha_numeric_key = to_alpha_numeric(user)
    # always use the default first
    updated_avatars = default_avatars.copy()
    # update with the user input
    updated_avatars.update(avatars)
    # correct the keys to be alpha numeric
    updated_avatars = {
        to_alpha_numeric(key): value for key, value in updated_avatars.items()
    }

    # now lookup the avatar
    avatar = updated_avatars.get(alpha_numeric_key, avatar)
    if isinstance(avatar, str):
        avatar = avatar.format(dist_path=CDN_DIST)
    return avatar


def build_avatar_pane(
    avatar: Any, css_classes: list[str], width: int = 15, height: int = 15
) -> Image | HTML:
    avatar_params = {
        "css_classes": css_classes,
        "width": width,
        "height": height,
    }
    if isinstance(avatar, Viewable):
        avatar_pane = avatar
        avatar_params["css_classes"] = (
            avatar_params.get("css_classes", []) + avatar_pane.css_classes
        )
        avatar_pane.param.update(avatar_params)
    elif not isinstance(avatar, (BytesIO, bytes)) and len(avatar) == 1:
        # single character
        avatar_pane = HTML(avatar, **avatar_params)
    else:
        try:
            avatar_pane = Image(avatar, **avatar_params)
        except ValueError:
            # likely an emoji
            avatar_pane = HTML(avatar, **avatar_params)
    return avatar_pane


def stream_to(obj, token, replace=False, object_panel=None):
    """
    Updates the message with the new token traversing the object to
    allow updating nested objects. When traversing a nested Panel
    the last object that supports rendering strings is updated, e.g.
    in a layout of `Column(Markdown(...), Image(...))` the Markdown
    pane is updated.

    Arguments
    ---------
    token: str
        The token to stream to the text pane.
    replace: bool (default=False)
        Whether to replace the existing text.
    """
    i = -1
    parent_panel = None
    attr = "object"
    if obj is None:
        obj = ""

    while not isinstance(obj, str) or isinstance(object_panel, ImageBase):
        object_panel = obj
        if hasattr(obj, "objects"):
            parent_panel = obj
            attr = "objects"
            obj = obj.objects[i]
            i = -1
        elif hasattr(obj, "object"):
            attr = "object"
            obj = obj.object
        elif hasattr(obj, "value"):
            attr = "value"
            obj = obj.value
        elif parent_panel is not None:
            obj = parent_panel
            parent_panel = None
            i -= 1
    contents = token if replace else obj + token
    setattr(object_panel, attr, contents)
    return object_panel


def get_obj_label(obj):
    """
    Get the label for the object; defaults to specified object name;
    if unspecified, defaults to the type name.
    """
    label = obj.name
    type_name = type(obj).__name__
    # If the name is just type + ID, simply use type
    # e.g. Column10241 -> Column
    if label.startswith(type_name) or not label:
        label = type_name
    return label


def serialize_recursively(
    obj: Any,
    prefix_with_viewable_label: bool = True,
    prefix_with_container_label: bool = True,
) -> str:
    """
    Recursively serialize the object to a string.
    """
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        content = tuple(
            serialize_recursively(
                o,
                prefix_with_viewable_label=prefix_with_viewable_label,
                prefix_with_container_label=prefix_with_container_label,
            )
            for o in obj
        )
        if prefix_with_container_label:
            if len(content) == 1:
                return f"{get_obj_label(obj)}({content[0]})"
            else:
                indented_content = indent(",\n".join(content), prefix=" " * 4)
                # outputs like:
                # Row(
                #   1,
                #   "str",
                # )
                return f"{get_obj_label(obj)}(\n{indented_content}\n)"
        else:
            # outputs like:
            # (1, "str")
            return f"({', '.join(content)})"

    string = obj
    if hasattr(obj, "value"):
        string = obj.value
    elif hasattr(obj, "object"):
        string = obj.object

    if hasattr(string, "decode") or isinstance(string, BytesIO):
        param.main.param.warning(
            f"Serializing byte-like objects are not supported yet; "
            f"using the label of the object as a placeholder for {obj}"
        )
        return get_obj_label(obj)

    if prefix_with_viewable_label and isinstance(obj, Viewable):
        label = get_obj_label(obj)
        string = f"{label}={string!r}"

    if not isinstance(string, str):
        string = str(string)

    return string



import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.chat import ChatMessage
from panel.tests.util import serve_component

pytestmark = pytest.mark.ui


def test_chat_message_dynamic_footer_objects(page):
    chat_msg = ChatMessage()
    serve_component(page, chat_msg)

    footer = page.get_by_text("It works!")
    expect(footer).to_have_count(0)

    chat_msg.footer_objects = ["It works!"]
    footer = page.get_by_text("It works!")
    expect(footer).to_have_count(1)


def test_chat_message_dynamic_header_objects(page):
    chat_msg = ChatMessage()
    serve_component(page, chat_msg)

    header = page.get_by_text("It works!")
    expect(header).to_have_count(0)

    chat_msg.header_objects = ["It works!"]
    header = page.get_by_text("It works!")
    expect(header).to_have_count(1)
